# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import contextlib
import hashlib
import json
import logging
import platform
from collections.abc import Iterable, Iterator
from datetime import timedelta
from functools import cached_property
from typing import Any, Literal

import attrs
import lance
import lancedb
import numpy as np
import pyarrow as pa
from lancedb import AsyncConnection, connect_async
from lancedb._lancedb import MergeResult
from lancedb.common import DATA, VECTOR_COLUMN_NAME
from lancedb.index import IndexConfig
from lancedb.merge import LanceMergeInsertBuilder
from lancedb.query import LanceQueryBuilder, LanceTakeQueryBuilder
from lancedb.query import Query as LanceQuery
from lancedb.table import IndexStatistics, TableStatistics, Tags
from lancedb.table import LanceTable as LanceLocalTable
from lancedb.table import Table as LanceTable
from lancedb.types import OnBadVectorsType
from pyarrow.fs import FileSystem, LocalFileSystem

# Python 3.10 compatibility
from typing_extensions import Never, override  # noqa: UP035
from yarl import URL

from geneva.checkpoint import (
    CheckpointStore,
    InMemoryCheckpointStore,
)
from geneva.db import Connection, connect
from geneva.query import GenevaQueryBuilder
from geneva.transformer import UDF, UDFArgType
from geneva.utils import status_updates

_LOG = logging.getLogger(__name__)


@attrs.define
class JobFuture:
    job_id: str

    def done(self, timeout: float | None = None) -> bool:
        raise NotImplementedError("JobFuture.done() must be implemented in subclasses")

    def result(self, timeout: float | None = None) -> Any:
        raise NotImplementedError(
            "JobFuture.result() must be implemented in subclasses"
        )

    def status(self, timeout: float | None = None) -> None:
        raise NotImplementedError(
            "JobFuture.status() must be implemented in subclasses"
        )


@attrs.define(order=True)
class TableReference:
    """
    Serializable reference to a Geneva Table.

    Used to pass through ray.remote calls
    """

    db_uri: str
    table_name: str
    version: int | None

    def open_checkpoint_store(self) -> CheckpointStore:
        """Open a Lance checkpoint store for this table."""
        try:
            return CheckpointStore.from_uri(str(URL(str(self.db_uri)) / "ckp"))
        except Exception:
            # Fallback to in-memory checkpoint store if Lance store fails
            return InMemoryCheckpointStore()

    def open_db(self) -> Connection:
        """Open a connection to the Lance database.
        Set read consistency interval to 0 for strongly consistent reads."""
        return connect(
            self.db_uri,
            checkpoint=self.open_checkpoint_store(),
            read_consistency_interval=timedelta(0),
        )

    async def open_db_async(self) -> AsyncConnection:
        """Open an async connection to the Lance database.
        This uses native lancedb AsyncConnection and doesn't support checkpoint store.
        """
        return await connect_async(
            self.db_uri,
            read_consistency_interval=timedelta(0),
        )

    def open(self) -> "Table":
        return self.open_db().open_table(self.table_name, version=self.version)


class Table(LanceTable):
    """Table in Geneva.

    A Table is a Lance dataset
    """

    def __init__(
        self,
        conn: Connection,
        name: str,
        *,
        version: int | None = None,
        storage_options: dict[str, str] | None = None,
        index_cache_size: int | None = None,
        **kwargs,
    ) -> None:
        self._conn_uri = conn.uri
        self._name = name

        self._conn = conn

        base_uri = URL(conn.uri)
        self._uri = str(base_uri / f"{name}.lance")
        self._version: int | None = version
        self._index_cache_size = index_cache_size
        self._storage_options = storage_options

        # Load table
        self._ltbl  # noqa

    def __repr__(self) -> str:
        return f"<Table {self._name}>"

    # TODO: This annotation sucks
    def __reduce__(self):  # noqa: ANN204
        return (self.__class__, (self._conn, self._name))

    def get_reference(self) -> TableReference:
        return TableReference(
            db_uri=self._conn.uri, table_name=self._name, version=self._version
        )

    def get_fragments(self) -> list[lance.LanceFragment]:
        return self.to_lance().get_fragments()

    @cached_property
    def _ltbl(self) -> lancedb.table.Table:
        # Open table (remote or local)
        if self._conn_uri.startswith("db://"):
            tbl = self._conn._connect.open_table(self._name)
        else:
            tbl = self._conn._connect.open_table(self.name)

        # Check out the specified version regardless of database type
        if self._version:
            tbl.checkout(self._version)

        return tbl

    @property
    def name(self) -> str:
        """Get the name of the table."""
        return self._name

    @property
    def version(self) -> int:
        """Get the current version of the table"""
        return self._ltbl.version

    @property
    def schema(self) -> pa.Schema:
        """The Arrow Schema of the Table."""
        return self._ltbl.schema

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def embedding_functions(self) -> Never:
        raise NotImplementedError("Embedding functions are not supported.")

    def add(
        self,
        data,
        mode: str = "append",
        on_bad_vectors: str = "error",
        fill_value: float = 0.0,
    ) -> None:
        self._ltbl.add(
            data,
            mode=mode,  # type: ignore[arg-type]
            on_bad_vectors=on_bad_vectors,  # type: ignore[arg-type]
            fill_value=fill_value,
        )

    def checkout(self, version: int) -> None:
        self._version = version
        self._ltbl.checkout(version)

    def checkout_latest(self) -> None:
        self._ltbl.checkout_latest()

    def add_columns(
        self, transforms: dict[str, str | UDF | tuple[UDF, list[str]]], *args, **kwargs
    ) -> None:
        """
        Add columns or UDF-based columns to the Geneva table.

        For UDF columns, this method validates that:
        - All input columns exist in the table schema
        - Column types are compatible with UDF type annotations (if present)
        - RecordBatch UDFs do not have input_columns defined

        This early validation helps catch configuration errors before job execution.

        Parameters
        ----------
        transforms : dict[str, str | UDF | tuple[UDF, list[str]]]
            The key is the column name to add and the value is a
            specification of the column type/value.

            * If the spec is a string, it is expected to be a datafusion
              sql expression. (e.g "cast(null as string)")
            * If the spec is a UDF, a virtual column is added with input
              columns inferred from the UDF's argument names.
            * If the spec is a tuple, the first element is a UDF and the
              second element is a list of input column names.

        Raises
        ------
        ValueError
            If UDF validation fails (missing columns, type mismatches, etc.)

        Warns
        -----
        UserWarning
            If type validation is skipped due to missing type annotations

        Examples
        --------
        >>> @udf(data_type=pa.int32())
        ... def double(a: int) -> int:
        ...     return a * 2
        >>> table.add_columns({"doubled": double})  # Validates 'a' column exists

        """
        # handle basic columns
        basic_cols = {k: v for k, v in transforms.items() if isinstance(v, str)}
        if len(basic_cols) > 0:
            self._ltbl.add_columns(basic_cols, *args)

        # handle UDF virtual columns
        udf_cols = {k: v for k, v in transforms.items() if not isinstance(v, str)}
        for k, v in udf_cols.items():
            if isinstance(v, UDF):
                # infer column names from udf arguments
                udf = v
                self._add_virtual_columns(
                    {k: udf}, *args, input_columns=udf.input_columns, **kwargs
                )
            else:
                # explicitly specify input columns
                (udf, cols) = v
                self._add_virtual_columns({k: udf}, *args, input_columns=cols, **kwargs)

    def _add_virtual_columns(
        self,
        mapping: dict[str, UDF],  # this breaks the non udf mapping
        *args,
        input_columns: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        This is an internal method and not intended to be called directly.

        Add udf based virtual columns to the Geneva table.
        """

        if len(mapping) != 1:
            raise ValueError("Only one UDF is supported for now.")

        _LOG.info("Adding column: udf=%s", mapping)
        col_name = next(iter(mapping))
        udf = mapping[col_name]

        if not isinstance(udf, UDF):
            # Stateful udfs are implemenated as Callable classses, and look
            # like partial functions here.  Instantiate to get the return
            # data_type annotations.
            udf = udf()

        # Validate input columns exist in table schema before adding the column
        self._validate_udf_input_columns(udf, input_columns)

        # Check for circular dependencies before adding the column
        cols_to_check = (
            input_columns if input_columns is not None else udf.input_columns
        )
        if (
            udf.arg_type != UDFArgType.RECORD_BATCH
            and cols_to_check
            and col_name in cols_to_check
        ):
            raise ValueError(
                f"UDF output column {col_name} is not allowed to be in"
                f" input {cols_to_check}"
            )

        self._ltbl.add_columns(pa.field(col_name, udf.data_type))
        self._configure_computed_column(col_name, udf, input_columns)

    def _validate_udf_input_columns(
        self, udf: UDF, input_columns: list[str] | None
    ) -> None:
        """
        Validate that UDF input columns exist in the table schema.

        This method delegates to the UDF's validate_against_schema() method
        for consolidated validation logic.

        Parameters
        ----------
        udf: UDF
            The UDF to validate
        input_columns: list[str] | None
            The input column names to validate

        Raises
        ------
        ValueError: If input columns don't exist in table schema or have type mismatches
        """
        # Delegate to UDF's consolidated validation method
        udf.validate_against_schema(self._ltbl.schema, input_columns)

    def refresh(self, where: str | None = None, src_version: int | None = None) -> None:
        """
        Refresh the specified materialized view.

        Parameters
        ----------
        where: str | None
            TODO: sql expression filter used to only backfill selected rows
        src_version: int | None
            Optional source table version to refresh from. If None (default),
            uses the latest version of the source table.
        """
        if where:
            raise NotImplementedError(
                "where clauses on materialized view refresh not implemented yet."
            )

        from geneva.runners.ray.pipeline import run_ray_copy_table

        run_ray_copy_table(
            self.get_reference(),
            self._conn._packager,
            self._conn._checkpoint_store,
            src_version=src_version,
        )
        self.checkout_latest()

    def backfill_async(
        self,
        col_name: str,
        *,
        udf: UDF | None = None,
        where: str | None = None,
        _enable_job_tracker_saves: bool = True,
        **kwargs,
    ) -> JobFuture:
        """
        Backfills the specified column asynchronously.

        Returns job future. Call .result() to wait for completion.

        Parameters
        ----------
        col_name: str
            Target column name to backfill
        udf: UDF | None
            Optionally override the UDF used to backfill the column.
        where: str | None
            SQL expression filter used select rows to apply backfills.
        concurrency: int
            (default = 8) This controls the number of processes that tasks run
            concurrently. For max throughput, ideally this is larger than the number
            of nodes in the k8s cluster.   This is the number of Ray actor processes
            are started.
        intra_applier_concurrency: int
            (default = 1) This controls the number of threads used to execute tasks
            within a process. Multiplying this times `concurrency` roughly corresponds
            to the number of cpu's being used.
        commit_granularity: int | None
            (default = 64) Show a partial result everytime this number of fragments
            are completed. If None, the entire result is committed at once.
        read_version: int | None
            (default = None) The version of the table to read from.  If None, the
            latest version is used.
        task_shuffle_diversity: int | None
            (default = 8) ??
        batch_size: int | None
            (default = 10240) The number of rows per batch when reading data from the
            table. If None, the default value is used.
        num_frags: int | None
            (default = None) The number of table fragments to process.  If None,
            process all fragments.
        _enable_job_tracker_saves: bool
            (default = False) Experimentally enable persistence of job metrics to the
            database. When disabled, metrics are tracked in-memory only.
        """

        from geneva.runners.ray.pipeline import (
            dispatch_run_ray_add_column,
            validate_backfill_args,
        )

        validate_backfill_args(self, col_name, udf)

        fut = dispatch_run_ray_add_column(
            self.get_reference(),
            col_name,
            udf=udf,
            where=where,
            enable_job_tracker_saves=_enable_job_tracker_saves,
            **kwargs,
        )
        return fut

    def backfill(
        self,
        col_name,
        *,
        udf: UDF | None = None,
        where: str | None = None,
        concurrency: int = 8,
        intra_applier_concurrency: int = 1,
        refresh_status_secs: float = 2.0,
        _enable_job_tracker_saves: bool = True,
        **kwargs,
    ) -> str:
        """
        Backfills the specified column.

        Returns job_id string

        Parameters
        ----------
        col_name: str
            Target column name to backfill
        udf: UDF | None
            Optionally override the UDF used to backfill the column.
        where: str | None
            SQL expression filter used select rows to apply backfills.
        concurrency: int
            (default = 8) This controls the number of processes that tasks run
            concurrently. For max throughput, ideally this is larger than the number
            of nodes in the k8s cluster.   This is the number of Ray actor processes
            are started.
        intra_applier_concurrency: int
            (default = 1) This controls the number of threads used to execute tasks
            within a process. Multiplying this times `concurrency` roughly corresponds
            to the number of cpu's being used.
        commit_granularity: int | None
            (default = 64) Show a partial result everytime this number of fragments
            are completed. If None, the entire result is committed at once.
        read_version: int | None
            (default = None) The version of the table to read from.  If None, the
            latest version is used.
        task_shuffle_diversity: int | None
            (default = 8) ??
        batch_size: int | None
            (default = 100) The number of rows per batch when reading data from the
            table. If None, the default value is used.  If 0, the batch will be the
            total number of rows from a fragment.
        num_frags: int | None
            (default = None) The number of table fragments to process.  If None,
            process all fragments.
        _enable_job_tracker_saves: bool
            (default = False) Experimentally enable persistence of job metrics to the
            database. When disabled, metrics are tracked in-memory only.
        """
        # Input validation
        from geneva.runners.ray.pipeline import validate_backfill_args

        validate_backfill_args(self, col_name, udf)

        # get cluster status
        from geneva.runners.ray.raycluster import ClusterStatus

        cs = ClusterStatus()
        try:
            with status_updates(cs.get_status, refresh_status_secs):
                # Kick off the job
                fut = self.backfill_async(
                    col_name,
                    udf=udf,
                    where=where,
                    concurrency=concurrency,
                    intra_applier_concurrency=intra_applier_concurrency,
                    _enable_job_tracker_saves=_enable_job_tracker_saves,
                    **kwargs,
                )

            while not fut.done(timeout=refresh_status_secs):
                # wait for the backfill to complete, updating statuses
                cs.get_status()
                fut.status()

            cs.get_status()
            fut.status()

            # Check for errors - this will raise if the job failed
            fut.result()

            # updates came from an external writer, so get the latest version.
            self._ltbl.checkout_latest()
            return fut.job_id
        finally:
            with contextlib.suppress(Exception):
                cs.close()

    def alter_columns(self, *alterations: dict[str, Any], **kwargs) -> None:
        """
        Alter columns in the table.  This can change the computed columns' udf

        Parameters
        ----------
        alterations:  Iterable[dict[str, Any]]
            This is a list of alterations to apply to the table.


        Example:
            >>> alter_columns({ "path": "col1", "udf": col1_udf_v2, })`
            >>> t.alter_columns(b
            ...     { "path": "col1", "udf": col1_udf_v2, },
            ...     { "path": "col2", "udf": col2_udf})

        """
        basic_column_alterations = []
        for alter in alterations:
            if "path" not in alter:
                raise ValueError("path is required to alter computed column's udf")

            if "virtual_column" in alter:  # deprecated
                udf = alter.get("virtual_column")
                if not isinstance(udf, UDF):
                    raise ValueError("virtual_column must be a UDF")
                _LOG.warning(
                    "alter_columns 'virtual_column' is deprecated, use 'udf' instead."
                )
            elif "udf" in alter:
                udf = alter.get("udf")
                if not isinstance(udf, UDF):
                    raise ValueError("udf must be a UDF")
            else:
                basic_column_alterations.append(alter)
                continue

            col_name = alter["path"]

            input_cols = alter.get("input_columns", None)
            if input_cols is None:
                input_cols = udf.input_columns

            self._configure_computed_column(col_name, udf, input_cols)

        if len(basic_column_alterations) > 0:
            self._ltbl.alter_columns(*basic_column_alterations)

    def _configure_computed_column(
        self,
        col_name: str,
        udf: UDF,
        input_cols: list[str] | None,
    ) -> None:
        """
        Configure a column to be a computed column for the given UDF.

        This procedure includes:
        - Packaging the UDF
        - Uploading the UDF to the dataset
        - Updating the field metadata to include the UDF information

        Note that the column should already exist on the table.
        """
        # record batch udf's don't specify inputs
        if (
            udf.arg_type != UDFArgType.RECORD_BATCH
            and udf.input_columns
            and col_name in udf.input_columns
        ):
            raise ValueError(
                f"UDF output column {col_name} is not allowed to be in"
                f" input {udf.input_columns}"
            )

        udf_spec = self._conn._packager.marshal(udf)

        # upload the UDF to the dataset URL
        if not isinstance(self._ltbl, LanceLocalTable):
            raise TypeError(
                "adding udf column is currently only supported for local tables"
            )

        # upload the packaged UDF to some location inside the dataset:
        ds = self.to_lance()
        fs, root_uri = FileSystem.from_uri(ds.uri)
        checksum = hashlib.sha256(udf_spec.udf_payload).hexdigest()
        udf_location = f"_udfs/{checksum}"

        # TODO -- only upload the UDF if it doesn't exist
        if isinstance(fs, LocalFileSystem):
            # Object storage filesystems like GCS and S3 will create the directory
            # automatically, but local filesystem will not, so we create explicitly
            fs.create_dir(f"{root_uri}/_udfs")

        with fs.open_output_stream(f"{root_uri}/{udf_location}") as f:
            f.write(udf_spec.udf_payload)

        # TODO rename this from virtual_column to computed column
        field_metadata = udf.field_metadata | {
            "virtual_column": "true",
            "virtual_column.udf_backend": udf_spec.backend,
            "virtual_column.udf_name": udf_spec.name,
            "virtual_column.udf": "_udfs/" + checksum,
            "virtual_column.udf_inputs": json.dumps(input_cols),
            "virtual_column.platform.system": platform.system(),
            "virtual_column.platform.arch": platform.machine(),
            "virtual_column.platform.python_version": platform.python_version(),
        }

        # Add the column metadata:
        self._ltbl.replace_field_metadata(col_name, field_metadata)

    def create_index(
        self,
        metric: str = "L2",
        num_partitions: int | None = None,
        num_sub_vectors: int | None = None,
        vector_column_name: str = VECTOR_COLUMN_NAME,
        replace: bool = True,
        accelerator=None,
        index_cache_size=None,
        *,
        index_type: Literal[
            "IVF_FLAT",
            "IVF_PQ",
            "IVF_HNSW_SQ",
            "IVF_HNSW_PQ",
        ] = "IVF_PQ",
        num_bits: int = 8,
        max_iterations: int = 50,
        sample_rate: int = 256,
        m: int = 20,
        ef_construction: int = 300,
    ) -> None:
        """Create Vector Index"""
        self._ltbl.create_index(
            metric,
            num_partitions or 256,
            num_sub_vectors or 96,
            vector_column_name,
            replace,
            accelerator,
            index_cache_size,
            index_type=index_type,
            num_bits=num_bits,
            max_iterations=max_iterations,
            sample_rate=sample_rate,
            m=m,
            ef_construction=ef_construction,
        )

    @override
    def create_fts_index(
        self,
        field_names: str | list[str],
        *,
        ordering_field_names: str | list[str] | None = None,
        replace: bool = False,
        writer_heap_size: int | None = None,
        tokenizer_name: str | None = None,
        with_position: bool = True,
        base_tokenizer: Literal["simple", "raw", "whitespace"] = "simple",
        language: str = "English",
        max_token_length: int | None = 40,
        lower_case: bool = True,
        stem: bool = False,
        remove_stop_words: bool = False,
        ascii_folding: bool = False,
        **_kwargs,
    ) -> None:
        self._ltbl.create_fts_index(
            field_names,
            ordering_field_names=ordering_field_names,
            replace=replace,
            writer_heap_size=writer_heap_size,
            tokenizer_name=tokenizer_name,
            with_position=with_position,
            base_tokenizer=base_tokenizer,
            language=language,
            max_token_length=max_token_length,
            lower_case=lower_case,
            stem=stem,
            remove_stop_words=remove_stop_words,
            ascii_folding=ascii_folding,
            use_tantivy=False,
        )

    @override
    def create_scalar_index(
        self,
        column: str,
        *,
        replace: bool = True,
        index_type: Literal["BTREE", "BITMAP", "LABEL_LIST"] = "BTREE",
    ) -> None:
        self._ltbl.create_scalar_index(
            column,
            replace=replace,
            index_type=index_type,
        )

    @override
    def _do_merge(
        self,
        merge: LanceMergeInsertBuilder,
        new_data: DATA,
        on_bad_vectors: OnBadVectorsType,
        fill_value: float,
    ) -> MergeResult:
        return self._ltbl._do_merge(merge, new_data, on_bad_vectors, fill_value)

    @override
    def _execute_query(
        self,
        query: LanceQuery,
        batch_size: int | None = None,
    ) -> pa.RecordBatchReader:
        return self._ltbl._execute_query(query, batch_size=batch_size)

    def list_versions(self) -> list[dict[str, Any]]:
        return self._ltbl.list_versions()

    @override
    def cleanup_old_versions(
        self,
        older_than: timedelta | None = None,
        *,
        delete_unverified=False,
    ) -> Any:  # lance.CleanupStats not available in type stubs
        return self._ltbl.cleanup_old_versions(
            older_than,
            delete_unverified=delete_unverified,
        )

    def to_batches(self, batch_size: int | None = None) -> Iterator[pa.RecordBatch]:
        from .query import Query

        if isinstance(self._ltbl, Query):
            return self._ltbl.to_batches(batch_size)  # type: ignore[attr-defined]
        return self.to_lance().to_batches(batch_size)  # type: ignore[arg-type]

    """This is the signature for the standard LanceDB table.search call"""

    def search(  # type: ignore[override]
        self,
        query: list | pa.Array | pa.ChunkedArray | np.ndarray | None = None,
        vector_column_name: str | None = None,
        query_type: Literal["vector", "fts", "hybrid", "auto"] = "auto",
        ordering_field_name: str | None = None,
        fts_columns: str | list[str] | None = None,
    ) -> GenevaQueryBuilder | LanceQueryBuilder:
        if query is None:
            return GenevaQueryBuilder(self)
        else:
            return self._ltbl.search(
                query, vector_column_name, query_type, ordering_field_name, fts_columns
            )

    @override
    def drop_columns(self, columns: Iterable[str]) -> None:
        self._ltbl.drop_columns(columns)

    @override
    def to_arrow(self) -> pa.Table:
        return self._ltbl.to_arrow()

    @override
    def count_rows(self, filter: str | None = None) -> int:
        return self._ltbl.count_rows(filter)

    @override
    def update(
        self,
        where: str | None = None,
        values: dict | None = None,
        *,
        values_sql: dict[str, str] | None = None,
    ) -> None:
        self._ltbl.update(where, values, values_sql=values_sql)

    @override
    def delete(self, where: str) -> None:
        self._ltbl.delete(where)

    @override
    def list_indices(self) -> Iterable[IndexConfig]:
        return self._ltbl.list_indices()

    @override
    def index_stats(self, index_name: str) -> IndexStatistics | None:
        return self._ltbl.index_stats(index_name)

    @override
    def optimize(
        self,
        *,
        cleanup_older_than: timedelta | None = None,
        delete_unverified: bool = False,
    ) -> None:
        return self._ltbl.optimize(
            cleanup_older_than=cleanup_older_than,
            delete_unverified=delete_unverified,
        )

    @override
    def compact_files(self) -> None:
        self._ltbl.compact_files()

    @override
    def restore(self, *args, **kwargs) -> None:
        self._ltbl.restore(*args, **kwargs)

    # TODO: This annotation sucks
    def take_blobs(self, indices: list[int] | pa.Array, column: str):  # noqa: ANN201
        return self.to_lance().take_blobs(column, indices)  # type: ignore[arg-type]

    def to_lance(self) -> lance.LanceDataset:
        return self._ltbl.to_lance()  # type: ignore[attr-defined]

    def uses_v2_manifest_paths(self) -> bool:
        return self._ltbl.uses_v2_manifest_paths()

    def migrate_v2_manifest_paths(self) -> None:
        return self._ltbl.migrate_v2_manifest_paths()

    def _analyze_plan(self, query: LanceQuery) -> str:
        return self._ltbl._analyze_plan(query)

    def _explain_plan(self, query: LanceQuery, verbose: bool | None = False) -> str:
        return self._ltbl._explain_plan(query, verbose=verbose)

    def stats(self) -> TableStatistics:
        return self._ltbl.stats()

    @property
    def tags(self) -> Tags:
        return self._ltbl.tags

    def take_offsets(self, offsets: list[int]) -> LanceTakeQueryBuilder:
        return self._ltbl.take_offsets(offsets)

    def take_row_ids(self, row_ids: list[int]) -> LanceTakeQueryBuilder:
        return self._ltbl.take_row_ids(row_ids)

    def get_errors(
        self,
        job_id: str | None = None,
        column_name: str | None = None,
        error_type: str | None = None,
    ) -> list[Any]:
        """Get error records for this table.

        Parameters
        ----------
        job_id : str, optional
            Filter errors by job ID
        column_name : str, optional
            Filter errors by column name
        error_type : str, optional
            Filter errors by exception type

        Returns
        -------
        list[ErrorRecord]
            List of error records matching the filters

        Examples
        --------
        >>> # Get all errors for this table
        >>> errors = table.get_errors()
        >>>
        >>> # Get errors for a specific job
        >>> errors = table.get_errors(job_id="abc123")
        >>>
        >>> # Get errors for a specific column
        >>> errors = table.get_errors(column_name="my_column")
        """
        from geneva.debug.error_store import ErrorStore

        error_store = ErrorStore(self._conn)
        return error_store.get_errors(
            job_id=job_id,
            table_name=self._name,
            column_name=column_name,
            error_type=error_type,
        )

    def get_failed_row_addresses(self, job_id: str, column_name: str) -> list[int]:
        """Get row addresses for all failed rows in a job.

        Parameters
        ----------
        job_id : str
            Job ID to query
        column_name : str
            Column name to filter by

        Returns
        -------
        list[int]
            List of row addresses that failed

        Examples
        --------
        >>> # Get failed row addresses
        >>> failed_rows = table.get_failed_row_addresses(
        ...     job_id="abc123", column_name="my_col"
        ... )
        >>>
        >>> # Retry processing only failed rows
        >>> row_ids = ','.join(map(str, failed_rows))
        >>> table.backfill("my_col", where=f"_rowaddr IN ({row_ids})")
        """
        from geneva.debug.error_store import ErrorStore

        error_store = ErrorStore(self._conn)
        return error_store.get_failed_row_addresses(
            job_id=job_id, column_name=column_name
        )

    @override
    def _output_schema(self, query: LanceQuery) -> pa.Schema:
        return self._ltbl._output_schema(query)



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import tempfile

import pyarrow as pa
import pytest

from geneva.checkpoint import (
    CheckpointConfig,
    CheckpointStore,
    InMemoryCheckpointStore,
    LanceCheckpointStore,
    LanceSessionizedCheckpointStore,
)


@pytest.mark.parametrize(
    "store",
    [
        InMemoryCheckpointStore(),
        LanceCheckpointStore(f"{tempfile.mkdtemp()}/new_dir"),
        LanceSessionizedCheckpointStore(f"{tempfile.mkdtemp()}/new_dir2"),
    ],
)
def test_checkpoint(store: CheckpointStore) -> None:
    store["key"] = pa.RecordBatch.from_pydict({"a": [1, 2, 3]})
    assert "key" in store
    assert "key" in list(store.list_keys())
    assert store["key"].to_pydict() == {"a": [1, 2, 3]}


def test_default_ckp_store() -> None:
    store = CheckpointConfig(mode="tempfile").make()
    assert isinstance(store, LanceCheckpointStore)
    assert store.root.startswith(tempfile.gettempdir())



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import copy
import platform
import sys

import ray

from geneva.cluster import GenevaClusterType, K8sConfigMethod
from geneva.cluster.mgr import (
    GenevaCluster,
    HeadGroupConfig,
    KubeRayConfig,
    WorkerGroupConfig,
)
from geneva.runners.ray.raycluster import (
    GENEVA_RAY_CPU_NODE,
    GENEVA_RAY_GPU_NODE,
    GENEVA_RAY_HEAD_NODE,
    get_ray_image,
)


class HeadGroupBuilder:
    """Builder for HeadGroupConfig. This can be used to build the configuration for
    a Ray head group with reasonable defaults. See GenevaClusterBuilder for examples"""

    def __init__(self) -> None:
        self._image: str = _default_image()
        self._num_cpus: int = 2
        self._memory: str = "4Gi"
        self._num_gpus: int = 0
        self._service_account: str = "geneva-service-account"
        self._node_selector: dict[str, str] = {GENEVA_RAY_HEAD_NODE: "true"}
        self._labels: dict[str, str] = {}
        self._tolerations: list[dict[str, str]] = []

    def image(self, image: str) -> "HeadGroupBuilder":
        """Set the container image."""
        self._image = image
        return self

    def cpus(self, cpus: int) -> "HeadGroupBuilder":
        """Set the number of CPUs."""
        self._num_cpus = cpus
        return self

    def memory(self, memory: str) -> "HeadGroupBuilder":
        """Set the memory allocation (e.g., '4Gi', '8Gi')."""
        self._memory = memory
        return self

    def gpus(self, gpus: int) -> "HeadGroupBuilder":
        """Set the number of GPUs."""
        self._num_gpus = gpus
        return self

    def service_account(self, service_account: str) -> "HeadGroupBuilder":
        """Set the Kubernetes service account."""
        self._service_account = service_account
        return self

    def node_selector(self, node_selector: dict[str, str]) -> "HeadGroupBuilder":
        """Set the node selector for pod placement."""
        self._node_selector = node_selector.copy()
        return self

    def labels(self, labels: dict[str, str]) -> "HeadGroupBuilder":
        """Set the pod labels."""
        self._labels = labels.copy()
        return self

    def tolerations(self, tolerations: list[dict[str, str]]) -> "HeadGroupBuilder":
        """Set the pod tolerations."""
        self._tolerations = tolerations.copy()
        return self

    def add_label(self, key: str, value: str) -> "HeadGroupBuilder":
        """Add a single label."""
        self._labels[key] = value
        return self

    def add_toleration(
        self, key: str, operator: str = "Equal", value: str = "", effect: str = ""
    ) -> "HeadGroupBuilder":
        """Add a single toleration."""
        toleration = {"key": key, "operator": operator}
        if value:
            toleration["value"] = value
        if effect:
            toleration["effect"] = effect
        self._tolerations.append(toleration)
        return self

    def build(self) -> HeadGroupConfig:
        """Build the HeadGroupConfig."""
        return HeadGroupConfig(
            service_account=self._service_account,
            num_cpus=self._num_cpus,
            memory=self._memory,
            image=self._image,
            num_gpus=self._num_gpus,
            node_selector=self._node_selector,
            labels=self._labels,
            tolerations=self._tolerations,
        )

    @classmethod
    def create(cls) -> "HeadGroupBuilder":
        """Create a new head group builder with defaults."""
        return cls()

    @classmethod
    def cpu_head(cls, cpus: int = 2, memory: str = "4Gi") -> "HeadGroupBuilder":
        """Create a CPU-only head group."""
        return cls().cpus(cpus).memory(memory).gpus(0)


class WorkerGroupBuilder:
    """Builder for WorkerGroupConfig. This can be used to build the configuration for
    a Ray worker group with reasonable defaults. See GenevaClusterBuilder for
    examples"""

    def __init__(self) -> None:
        self._image: str = _default_image()
        self._num_cpus: int = 4
        self._memory: str = "8Gi"
        self._num_gpus: int = 0
        self._service_account: str = "geneva-service-account"
        self._node_selector: dict[str, str] = {GENEVA_RAY_CPU_NODE: "true"}
        self._labels: dict[str, str] = {}
        self._tolerations: list[dict[str, str]] = []

    def image(self, image: str) -> "WorkerGroupBuilder":
        """Set the container image."""
        self._image = image
        return self

    def cpus(self, cpus: int) -> "WorkerGroupBuilder":
        """Set the number of CPUs."""
        self._num_cpus = cpus
        return self

    def memory(self, memory: str) -> "WorkerGroupBuilder":
        """Set the memory allocation (e.g., '8Gi', '16Gi')."""
        self._memory = memory
        return self

    def gpus(self, gpus: int) -> "WorkerGroupBuilder":
        """Set the number of GPUs."""
        self._num_gpus = gpus
        return self

    def service_account(self, service_account: str) -> "WorkerGroupBuilder":
        """Set the Kubernetes service account."""
        self._service_account = service_account
        return self

    def node_selector(self, node_selector: dict[str, str]) -> "WorkerGroupBuilder":
        """Set the node selector for pod placement."""
        self._node_selector = node_selector.copy()
        return self

    def labels(self, labels: dict[str, str]) -> "WorkerGroupBuilder":
        """Set the pod labels."""
        self._labels = labels.copy()
        return self

    def tolerations(self, tolerations: list[dict[str, str]]) -> "WorkerGroupBuilder":
        """Set the pod tolerations."""
        self._tolerations = tolerations.copy()
        return self

    def add_label(self, key: str, value: str) -> "WorkerGroupBuilder":
        """Add a single label."""
        self._labels[key] = value
        return self

    def add_toleration(
        self, key: str, operator: str = "Equal", value: str = "", effect: str = ""
    ) -> "WorkerGroupBuilder":
        """Add a single toleration."""
        toleration = {"key": key, "operator": operator}
        if value:
            toleration["value"] = value
        if effect:
            toleration["effect"] = effect
        self._tolerations.append(toleration)
        return self

    def build(self) -> WorkerGroupConfig:
        """Build the WorkerGroupConfig."""
        # modify default image to enable GPU if GPUs are requested
        if self._image == _default_image() and self._num_gpus > 0:
            self._image = _default_image(gpu=True)

        return WorkerGroupConfig(
            service_account=self._service_account,
            num_cpus=self._num_cpus,
            memory=self._memory,
            image=self._image,
            num_gpus=self._num_gpus,
            node_selector=self._node_selector,
            labels=self._labels,
            tolerations=self._tolerations,
        )

    @classmethod
    def create(cls) -> "WorkerGroupBuilder":
        """Create a new worker group builder with defaults."""
        return cls()

    @classmethod
    def cpu_worker(cls, cpus: int = 4, memory: str = "8Gi") -> "WorkerGroupBuilder":
        """Create a CPU worker group."""
        return cls().cpus(cpus).memory(memory).gpus(0)

    @classmethod
    def gpu_worker(
        cls, cpus: int = 8, memory: str = "16Gi", gpus: int = 1
    ) -> "WorkerGroupBuilder":
        """Create a GPU worker group."""
        return (
            cls()
            .cpus(cpus)
            .node_selector({GENEVA_RAY_GPU_NODE: "true"})
            .memory(memory)
            .gpus(gpus)
        )


class GenevaClusterBuilder:
    """Fluent builder for GenevaCluster. `name` is required, all optional
    fields will use defaults.
        example usage:
         >>> GenevaClusterBuilder
         >>>    .name("my-cluster")
         >>>    .head_group_builder(HeadGroupBuilder().cpus(8).memory("16Gi"))
         >>>    .add_worker_group(WorkerGroupBuilder()
         >>>                         .gpu_worker(memory="8Gi", gpus=1))
         >>>    .build()
    """

    def __init__(self) -> None:
        self._name: str | None = None
        self._cluster_type: GenevaClusterType = GenevaClusterType.KUBE_RAY
        self._namespace: str = "geneva"
        self._config_method: K8sConfigMethod = K8sConfigMethod.LOCAL
        self._use_portforwarding: bool = True
        self._aws_region: str | None = None
        self._aws_role_name: str | None = None
        self._ray_init_kwargs: dict = {}

        # Head group defaults
        self._head_image: str = _default_image()
        self._head_cpus: int = 2
        self._head_memory: str = "4Gi"
        self._head_gpus: int = 0
        self._head_service_account: str = "geneva-service-account"
        self._head_node_selector: dict[str, str] = {}
        self._head_labels: dict[str, str] = {}
        self._head_tolerations: list[dict[str, str]] = []

        # Worker group defaults
        self._worker_groups: list[WorkerGroupConfig] = []

    def name(self, name: str) -> "GenevaClusterBuilder":
        """Set the cluster name."""
        self._name = name
        return self

    def cluster_type(self, cluster_type: GenevaClusterType) -> "GenevaClusterBuilder":
        """Set the cluster type."""
        self._cluster_type = cluster_type
        return self

    def namespace(self, namespace: str) -> "GenevaClusterBuilder":
        """Set the Kubernetes namespace."""
        self._namespace = namespace
        return self

    def config_method(self, method: K8sConfigMethod) -> "GenevaClusterBuilder":
        """Set the Kubernetes config method."""
        self._config_method = method
        return self

    def portforwarding(self, enabled: bool = True) -> "GenevaClusterBuilder":
        """Enable or disable port forwarding."""
        self._use_portforwarding = enabled
        return self

    def aws_config(
        self, region: str | None = None, role_name: str | None = None
    ) -> "GenevaClusterBuilder":
        """Configure AWS settings."""
        self._aws_region = region
        self._aws_role_name = role_name
        return self

    def ray_init_kwargs(self, kwargs: dict) -> "GenevaClusterBuilder":
        """
        Set arbitrary kwargs to pass to ray.init() when starting the cluster.

        Commonly used for runtime_env configuration with conda or pip
        dependencies.

        Example:
            >>> builder.ray_init_kwargs({
            ...     "runtime_env": {
            ...         "conda": {
            ...             "channels": ["conda-forge"],
            ...             "dependencies": [
            ...                 "python=3.10", "ffmpeg<8", "torchvision=0.22.1"
            ...             ]
            ...         },
            ...         "config": {"eager_install": True}
            ...     }
            ... })

        WARNING: This accepts arbitrary kwargs without validation allowing for injection
        of fault or potentially malicious configuration. Use with caution.
        """
        self._ray_init_kwargs = copy.deepcopy(kwargs)
        return self

    # Head group configuration
    def head_group(
        self,
        *,
        image: str | None = None,
        cpus: int | None = None,
        memory: str | None = None,
        gpus: int | None = None,
        service_account: str | None = None,
        node_selector: dict[str, str] | None = None,
        labels: dict[str, str] | None = None,
        tolerations: list[dict[str, str]] | None = None,
    ) -> "GenevaClusterBuilder":
        """Configure the head group with optional parameters."""
        if image is not None:
            self._head_image = image
        if cpus is not None:
            self._head_cpus = cpus
        if memory is not None:
            self._head_memory = memory
        if gpus is not None:
            self._head_gpus = gpus
        if service_account is not None:
            self._head_service_account = service_account
        if node_selector is not None:
            self._head_node_selector = node_selector
        if labels is not None:
            self._head_labels = labels
        if tolerations is not None:
            self._head_tolerations = tolerations
        return self

    def head_group_builder(self, builder: HeadGroupBuilder) -> "GenevaClusterBuilder":
        """Configure the head group using a HeadGroupBuilder."""
        head_config = builder.build()
        self._head_image = head_config.image
        self._head_cpus = head_config.num_cpus
        self._head_memory = head_config.memory
        self._head_gpus = head_config.num_gpus
        self._head_service_account = head_config.service_account
        self._head_node_selector = head_config.node_selector
        self._head_labels = head_config.labels
        self._head_tolerations = head_config.tolerations
        return self

    def add_cpu_worker_group(
        self,
        *,
        image: str | None = None,
        cpus: int = 4,
        memory: str = "8Gi",
        service_account: str | None = None,
        node_selector: dict[str, str] | None = None,
        labels: dict[str, str] | None = None,
        tolerations: list[dict[str, str]] | None = None,
    ) -> "GenevaClusterBuilder":
        """Add a CPU worker group."""
        worker = WorkerGroupConfig(
            image=image or self._head_image,
            num_cpus=cpus,
            memory=memory,
            num_gpus=0,
            service_account=service_account or "geneva-service-account",
            node_selector=node_selector or {GENEVA_RAY_CPU_NODE: "true"},
            labels=labels or {},
            tolerations=tolerations or [],
        )
        self._worker_groups.append(worker)
        return self

    def add_gpu_worker_group(
        self,
        *,
        image: str | None = None,
        cpus: int = 4,
        memory: str = "8Gi",
        gpus: int = 1,
        service_account: str | None = None,
        node_selector: dict[str, str] | None = None,
        labels: dict[str, str] | None = None,
        tolerations: list[dict[str, str]] | None = None,
    ) -> "GenevaClusterBuilder":
        """Add a GPU worker group."""
        worker = WorkerGroupConfig(
            image=image or self._head_image,
            num_cpus=cpus,
            memory=memory,
            num_gpus=gpus,
            service_account=service_account or "geneva-service-account",
            node_selector=node_selector or {GENEVA_RAY_GPU_NODE: "true"},
            labels=labels or {},
            tolerations=tolerations or [],
        )
        self._worker_groups.append(worker)
        return self

    def add_worker_group(self, builder: WorkerGroupBuilder) -> "GenevaClusterBuilder":
        """Add a worker group using a WorkerGroupBuilder."""
        worker = builder.build()
        self._worker_groups.append(worker)
        return self

    def build(self) -> GenevaCluster:
        """Build the GenevaCluster with the configured settings."""
        if self._name is None:
            raise ValueError("Cluster name is required. Use .name() to set it.")

        # Build head group config
        head_group = HeadGroupConfig(
            service_account=self._head_service_account,
            num_cpus=self._head_cpus,
            memory=self._head_memory,
            image=self._head_image,
            num_gpus=self._head_gpus,
            node_selector=self._head_node_selector,
            labels=self._head_labels,
            tolerations=self._head_tolerations,
        )

        # Use worker group configs directly (already WorkerGroupConfig objects)
        worker_groups = self._worker_groups.copy()

        # If no worker groups were explicitly added, add a default CPU worker
        if not worker_groups:
            worker_groups.append(
                WorkerGroupConfig(
                    service_account=self._head_service_account,
                    num_cpus=4,
                    memory="8Gi",
                    image=_default_image(False),
                    num_gpus=0,
                    node_selector={GENEVA_RAY_CPU_NODE: "true"},
                    labels={},
                    tolerations=[],
                )
            )

        supported_cluster_types = [GenevaClusterType.KUBE_RAY]
        if self._cluster_type not in supported_cluster_types:
            raise ValueError(
                f"cluster_type must be one of "
                f"{[e.name for e in supported_cluster_types]}"
            )

        kuberay_config = KubeRayConfig(
            namespace=self._namespace,
            head_group=head_group,
            worker_groups=worker_groups,
            config_method=self._config_method,
            use_portforwarding=self._use_portforwarding,
            aws_region=self._aws_region,
            aws_role_name=self._aws_role_name,
            ray_init_kwargs=self._ray_init_kwargs,
        )

        # Build and return the cluster
        return GenevaCluster(
            cluster_type=self._cluster_type,
            name=self._name,
            kuberay=kuberay_config,
        )

    @classmethod
    def create(cls, name: str) -> "GenevaClusterBuilder":
        """Create a new builder with the given cluster name."""
        return cls().name(name)

    @classmethod
    def local_cpu_cluster(
        cls, name: str, namespace: str = "geneva"
    ) -> "GenevaClusterBuilder":
        """Create a builder for a local CPU-only cluster with good defaults."""
        return (
            cls()
            .name(name)
            .namespace(namespace)
            .config_method(K8sConfigMethod.LOCAL)
            .portforwarding(True)
            .head_group(cpus=2, memory="4Gi")
            .add_cpu_worker_group(cpus=4, memory="8Gi")
        )

    @classmethod
    def gpu_cluster(
        cls, name: str, namespace: str = "geneva"
    ) -> "GenevaClusterBuilder":
        """Create a builder for a GPU cluster with good defaults."""
        return (
            cls()
            .name(name)
            .namespace(namespace)
            .config_method(K8sConfigMethod.IN_CLUSTER)
            .portforwarding(False)
            .head_group(cpus=2, memory="4Gi")
            .add_gpu_worker_group(cpus=4, memory="8Gi", gpus=1)
        )


def _default_image(gpu: bool = False) -> str:
    """Get the default Ray image"""
    ray_version = ray.__version__
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    return get_ray_image(
        ray_version,
        python_version,
        gpu=gpu,
        arm=platform.processor() in {"aarch64", "arm"},
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Simple CPU-only UDFs for image processing.

These UDFs have minimal dependencies and can run on any worker.
"""

import io

import pyarrow as pa

import geneva


@geneva.udf(version="0.1")
def file_size(image: bytes) -> int:
    """Compute the byte size of an image."""
    return len(image)


@geneva.udf(
    version="0.1",
    data_type=pa.struct(
        [pa.field("width", pa.int32()), pa.field("height", pa.int32())]
    ),
)
def dimensions(image: bytes) -> tuple[int, int]:
    """Extract image dimensions (width, height)."""
    from PIL import Image

    image_stream = io.BytesIO(image)
    img = Image.open(image_stream)
    return img.size



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
import json
import logging
import time

import attr
import attrs
import ray
from lancedb import AsyncConnection, AsyncTable
from lancedb.util import (
    value_to_sql,
)

from geneva.jobs.jobs import GENEVA_JOBS_TABLE_NAME
from geneva.table import TableReference
from geneva.utils import escape_sql_string

_LOG = logging.getLogger(__name__)


@ray.remote
@attrs.define
class JobTracker:
    """
    Centralized progress ledger for a job.
    - Metrics are arbitrary strings -> {n, total, done, desc}
    - All ops are atomic via the actor's mailbox.
    - DB saves can be disabled entirely with enable_saves=False
    """

    job_id: str

    table_ref: TableReference = attrs.field()

    metrics: dict[str, dict] = attrs.field(factory=dict)

    enable_saves: bool = attrs.field(default=True)

    min_time_between_updates_secs: float = attrs.field(default=5.0)

    # use async lancedb connection to avoid blocking Ray Tasks
    _db: AsyncConnection | None = attrs.field(init=False, default=None)

    _jobs_table: AsyncTable | None = attrs.field(init=False, default=None)

    _last_updated: float = attrs.field(init=False, default=-float("inf"))

    async def _upsert(
        self, name: str, *, total: int | None = None, desc: str | None = None
    ) -> None:
        """
        Given a metric 'name', ensure that the metric entry is present with specified
        values or initialized with 0s.
        This will flush the metrics to the underlying lance table if
        min_time_between_updates_secs is exceeded
        """
        m = self.metrics.get(name)
        if m is None:
            m = {"n": 0, "total": 0, "done": False, "desc": name}
            self.metrics[name] = m
        if total is not None:
            m["total"] = int(total)
        if desc is not None:
            m["desc"] = desc
        # auto-done if we're already at/over total
        if m["total"] and m["n"] >= m["total"]:
            m["done"] = True

        await self._save_with_throttle()

    # Generic metric API
    async def set_total(self, name: str, total: int) -> None:
        await self._upsert(name, total=total)

    async def set_desc(self, name: str, desc: str) -> None:
        await self._upsert(name, desc=desc)

    async def set(self, name: str, n: int) -> None:
        await self._upsert(name)
        m = self.metrics[name]
        m["n"] = int(n)
        if m["total"] and m["n"] >= m["total"]:
            m["done"] = True
            await self._save_with_throttle(force=True)

    async def increment(self, name: str, delta: int = 1) -> None:
        await self._upsert(name)
        m = self.metrics[name]
        m["n"] += int(delta)
        if m["total"] and m["n"] >= m["total"]:
            m["done"] = True
            await self._save_with_throttle(force=True)

    async def mark_done(self, name: str) -> None:
        await self._upsert(name)
        self.metrics[name]["done"] = True
        await self._save_with_throttle(force=True)

    # Read APIs
    async def get_progress(self, name: str) -> dict:
        await self._upsert(name)
        m = self.metrics[name]
        return {"n": m["n"], "total": m["total"], "done": m["done"], "desc": m["desc"]}

    def get_all(self) -> dict[str, dict]:
        # Shallow copy for safety
        return {k: dict(v) for k, v in self.metrics.items()}

    async def _save_with_throttle(self, force: bool = False) -> None:
        """Save metrics to DB, optionally bypassing throttle."""
        current_time = time.time()
        if (
            not force
            and self._last_updated + self.min_time_between_updates_secs > current_time
        ):
            # don't update metrics too frequently to avoid excessive DB writes
            return

        # bump last update and save metrics
        self._last_updated = current_time
        await self._save_metrics(attr.asdict(self)["metrics"])

    async def _save_metrics(self, _metrics: dict[str, dict]) -> None:
        if not self.enable_saves:
            return
        if not _metrics:
            return

        try:
            # lazy load db and table
            if not self._db:
                if not self.table_ref:
                    _LOG.error(
                        "JobTracker missing table_ref, metrics will not be saved"
                    )
                    return
                self._db = await self.table_ref.open_db_async()
                self._jobs_table = await self._db.open_table(GENEVA_JOBS_TABLE_NAME)

            updates = {
                "metrics": self._convert_metrics(_metrics),
            }
            updates_sql = {k: value_to_sql(v) for k, v in updates.items()}
            _LOG.debug(f"Saving metrics for job {self.job_id}: {updates_sql}")
            start = time.time()

            await self._jobs_table.update(
                where=f"job_id = '{escape_sql_string(self.job_id)}'",
                updates_sql=updates_sql,
            )
            _LOG.info(
                f"Saved metrics for job "
                f"{self.job_id} in {(time.time() - start) * 1000:.0f}ms"
            )

        except AttributeError as e:
            _LOG.warning(
                f"unable to save metrics. The execution manifest"
                f"may be using an older Geneva version than the client"
                f": {e}"
            )
        except Exception as e:
            _LOG.error(f"error saving metrics {e}")
            self._db = None  # force retry on next save

    def _convert_metrics(self, _metrics: dict[str, dict]) -> list[str]:
        """Convert metrics to list of json strings for storage"""
        m = []
        for name, v in _metrics.items():
            vm = dict(v)
            vm["name"] = name
            m.append(json.dumps(vm))
        return m

    def __str__(self) -> str:
        return f"JobTracker(job_id={self.job_id})"

    def __repr__(self) -> str:
        return str(self)



#!/usr/bin/env python
# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Upload manifest and add columns for video embeddings UDF to Geneva.

This script runs in the UDF's dependency environment (via uv run)
and uploads the manifest to the Geneva connection, then adds columns
to the specified table.

Usage:
    export GENEVA_TABLE_NAME=openvid_shared_abc123
    cd e2e/openvid/udfs/video-embeddings
    uv run python upload_manifest.py --bucket gs://bucket/path

Environment Variables:
    GENEVA_TABLE_NAME: Table name to add columns to (required)
    VIDEO_EMBEDDING_COL: Custom column name for video_embedding (default: "video_embedding")
"""

import logging
import os

from geneva.manifest.uploader_cli import main_upload_workflow
from manifest import create_manifest
from video_embedding_udf import VideoEmbedding

logging.basicConfig(level=logging.INFO)
_LOG = logging.getLogger(__name__)


def get_columns() -> dict[str, callable]:
    """
    Get column definitions for this UDF package.

    Column names can be customized via environment variables:
    - VIDEO_EMBEDDING_COL: Custom name for video_embedding column (default: "video_embedding_vjpea2")

    Returns:
        Dictionary mapping column names to UDF functions
    """
    return {
        os.getenv("VIDEO_EMBEDDING_COL", "video_embedding_vjepa2"): VideoEmbedding(),
    }


if __name__ == "__main__":
    main_upload_workflow(
        get_columns_fn=get_columns,
        create_manifest_fn=create_manifest,
        description="Upload video embeddings UDF manifest and add columns to table",
        default_manifest_name="video-embeddings-v1",
        logger=_LOG,
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import copy
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from geneva import connect
from geneva.cluster.builder import _default_image
from geneva.manifest.builder import GenevaManifestBuilder
from geneva.manifest.mgr import GenevaManifest


@pytest.mark.slow
def test_manifest_crud(tmp_path: Path) -> None:
    mock_uploader = MagicMock()
    mock_uploader.upload_dir = "/mock/upload/dir"
    mock_uploader._file_exists.return_value = False
    mock_uploader.upload.side_effect = lambda path: f"mock://{path.name}"

    geneva = connect(tmp_path)

    manifest_def = GenevaManifest(
        name="test-manifest-1",
        local_zip_output_dir=str(tmp_path),
        skip_site_packages=False,
        delete_local_zips=False,
        pip=["numpy", "pandas"],
        py_modules=["pyarrow"],
    )

    # upload and create
    geneva.define_manifest("test-manifest-1", manifest_def, uploader=mock_uploader)
    m = geneva.list_manifests()[0]
    _assert_manifest_eq(m.as_dict(), manifest_def.as_dict())

    upload_count = mock_uploader.upload.call_count
    assert upload_count >= 1, "files were not uploaded"

    # update - should update metadata and upload new artifacts
    manifest_def.skip_site_packages = True
    geneva.define_manifest("test-manifest-1", manifest_def, uploader=mock_uploader)
    manifests = geneva.list_manifests()
    assert len(manifests) == 1, "expected single manifest"
    m1 = manifests[0].as_dict()
    m2 = manifest_def.as_dict()
    assert m1["checksum"] != m2["checksum"], "checksum should change"
    _assert_manifest_eq(m1, m2)
    assert mock_uploader.upload.call_count >= upload_count, "files were not uploaded"

    # delete
    geneva.delete_manifest("test-manifest-1")
    assert geneva.list_manifests() == []


def _assert_manifest_eq(m1: dict, m2: dict) -> bool:
    m1 = copy.deepcopy(m1)
    m2 = copy.deepcopy(m2)
    # exclude transient fields from comparison
    for f in {"checksum", "zips"}:
        if f in m1:
            del m1[f]
        if f in m2:
            del m2[f]
    assert m1 == m2, "manifests should match"


class TestGenevaManifestBuilder:
    """Test suite for GenevaManifestBuilder."""

    def test_builder_minimal(self) -> None:
        """Test minimal manifest creation with builder."""
        manifest = GenevaManifestBuilder.create("test-manifest").build()

        assert manifest.name == "test-manifest"
        assert manifest.version is None
        assert manifest.pip == []
        assert manifest.py_modules == []
        assert manifest.head_image is None
        assert manifest.skip_site_packages is False
        assert manifest.delete_local_zips is False
        assert manifest.local_zip_output_dir is None

    def test_builder_fluent_api(self) -> None:
        """Test fluent API chaining."""
        manifest = (
            GenevaManifestBuilder()
            .name("test-manifest")
            .version("1.0.0")
            .pip(["numpy", "pandas"])
            .py_modules(["mymodule", "othermodule"])
            .head_image("custom:latest")
            .worker_image("custom:latest2")
            .skip_site_packages(True)
            .delete_local_zips(True)
            .local_zip_output_dir("/foo/tmp/zips")
            .build()
        )

        assert manifest.name == "test-manifest"
        assert manifest.version == "1.0.0"
        assert manifest.pip == ["numpy", "pandas"]
        assert manifest.py_modules == ["mymodule", "othermodule"]
        assert manifest.head_image == "custom:latest"
        assert manifest.worker_image == "custom:latest2"
        assert manifest.skip_site_packages is True
        assert manifest.delete_local_zips is True
        assert manifest.local_zip_output_dir == "/foo/tmp/zips"

    def test_builder_add_methods(self) -> None:
        """Test add_pip and add_py_module methods."""
        manifest = (
            GenevaManifestBuilder.create("test-manifest")
            .add_pip("numpy")
            .add_pip("pandas")
            .add_py_module("mymodule")
            .add_py_module("othermodule")
            .build()
        )

        assert manifest.pip == ["numpy", "pandas"]
        assert manifest.py_modules == ["mymodule", "othermodule"]

    def test_builder_list_immutability(self) -> None:
        """Test that lists are copied and not shared."""
        original_pip = ["numpy", "pandas"]
        original_modules = ["mymodule"]

        manifest1 = (
            GenevaManifestBuilder.create("manifest1")
            .pip(original_pip)
            .py_modules(original_modules)
            .build()
        )

        # Modify original lists
        original_pip.append("scipy")
        original_modules.append("othermodule")

        # Manifest should not be affected
        assert manifest1.pip == ["numpy", "pandas"]
        assert manifest1.py_modules == ["mymodule"]

        # Create another manifest with the same builder pattern
        manifest2 = (
            GenevaManifestBuilder.create("manifest2")
            .pip(original_pip)
            .py_modules(original_modules)
            .build()
        )

        # This manifest should have the modified lists
        assert manifest2.pip == ["numpy", "pandas", "scipy"]
        assert manifest2.py_modules == ["mymodule", "othermodule"]

    def test_builder_requires_name(self) -> None:
        """Test that builder requires a name."""
        with pytest.raises(ValueError, match="Manifest name is required"):
            GenevaManifestBuilder().build()

    def test_builder_class_method_create(self) -> None:
        """Test create class method."""
        manifest = GenevaManifestBuilder.create("created-manifest").build()

        assert manifest.name == "created-manifest"

    def test_builder_with_real_manifest_compatibility(self) -> None:
        """Test that builder creates manifests compatible with existing code."""
        # Create with builder
        built_manifest = (
            GenevaManifestBuilder.create("test-builder")
            .pip(["numpy", "pandas"])
            .py_modules(["pyarrow"])
            .skip_site_packages(True)
            .build()
        )

        # Create traditionally
        traditional_manifest = GenevaManifest(
            name="test-traditional",
            pip=["numpy", "pandas"],
            py_modules=["pyarrow"],
            skip_site_packages=True,
        )

        # Both should have same structure (excluding name and auto-generated fields)
        built_dict = built_manifest.as_dict()
        traditional_dict = traditional_manifest.as_dict()

        # Remove name and other varying fields for comparison
        for d in [built_dict, traditional_dict]:
            del d["name"]
            del d["checksum"]
            del d["created_at"]
            del d["created_by"]

        assert built_dict == traditional_dict

    def test_builder_version_handling(self) -> None:
        """Test version field handling."""
        manifest_with_version = (
            GenevaManifestBuilder.create("versioned").version("2.1.0").build()
        )

        manifest_without_version = GenevaManifestBuilder.create("unversioned").build()

        assert manifest_with_version.version == "2.1.0"
        assert manifest_without_version.version is None

    def test_builder_boolean_flags(self) -> None:
        """Test boolean flag methods."""
        # Test explicit True
        manifest1 = (
            GenevaManifestBuilder.create("test1")
            .skip_site_packages(True)
            .delete_local_zips(True)
            .build()
        )

        # Test explicit False
        manifest2 = (
            GenevaManifestBuilder.create("test2")
            .skip_site_packages(False)
            .delete_local_zips(False)
            .build()
        )

        # Test default (True)
        manifest3 = (
            GenevaManifestBuilder.create("test3")
            .skip_site_packages()
            .delete_local_zips()
            .build()
        )

        assert manifest1.skip_site_packages is True
        assert manifest1.delete_local_zips is True
        assert manifest2.skip_site_packages is False
        assert manifest2.delete_local_zips is False
        assert manifest3.skip_site_packages is True
        assert manifest3.delete_local_zips is True

    def test_default_image(self) -> None:
        m = (
            GenevaManifestBuilder.create("foo")
            .default_head_image()
            .default_worker_image()
            .build()
        )
        assert m.head_image == _default_image()
        assert m.worker_image == _default_image()



from pathlib import Path

import pyarrow as pa
import pytest
import ray
import ray.util.queue

from geneva.checkpoint import CheckpointStore
from geneva.runners.ray.writer import (
    _align_batches_to_physical_layout,
    _buffer_and_sort_batches,
)

pytestmark = pytest.mark.ray


def test_buffer_sort_and_align_batches_no_alignment_needed(
    tmp_path: Path,
) -> None:
    batch1 = pa.RecordBatch.from_arrays(
        [pa.array(range(16), type=pa.uint64()), pa.array(range(16), type=pa.uint64())],
        names=["data", "_rowaddr"],
    )

    batch2 = pa.RecordBatch.from_arrays(
        [
            pa.array(range(16, 32), type=pa.uint64()),
            pa.array(range(16, 32), type=pa.uint64()),
        ],
        names=["data", "_rowaddr"],
    )

    store = CheckpointStore.from_uri(tmp_path.as_posix())

    store["batch1"] = batch1
    store["batch2"] = batch2

    queue = ray.util.queue.Queue()
    queue.put(
        (
            16,
            "batch2",
        )
    )
    queue.put(
        (
            0,
            "batch1",
        )
    )

    assert list(
        _align_batches_to_physical_layout(
            32,
            32,
            0,
            _buffer_and_sort_batches(
                32,
                store,
                queue,
            ),
        )
    ) == [
        pa.RecordBatch.from_arrays(
            [
                pa.array(range(16), type=pa.uint64()),
                pa.array(range(16), type=pa.uint64()),
            ],
            names=["data", "_rowaddr"],
        ),
        pa.RecordBatch.from_arrays(
            [
                pa.array(range(16, 32), type=pa.uint64()),
                pa.array(range(16, 32), type=pa.uint64()),
            ],
            names=["data", "_rowaddr"],
        ),
    ]


def test_buffer_sort_and_align_batches_alignment_needed(
    tmp_path: Path,
) -> None:
    batch1 = pa.RecordBatch.from_arrays(
        [
            pa.array(range(0, 16, 2), type=pa.uint64()),
            pa.array(range(0, 16, 2), type=pa.uint64()),
        ],
        names=["data", "_rowaddr"],
    )

    batch2 = pa.RecordBatch.from_arrays(
        [
            pa.array(range(16, 32, 2), type=pa.uint64()),
            pa.array(range(16, 32, 2), type=pa.uint64()),
        ],
        names=["data", "_rowaddr"],
    )

    store = CheckpointStore.from_uri(tmp_path.as_posix())

    store["batch1"] = batch1
    store["batch2"] = batch2

    queue = ray.util.queue.Queue()
    queue.put(
        (
            8,
            "batch2",
        )
    )
    queue.put(
        (
            0,
            "batch1",
        )
    )

    batches = list(
        _align_batches_to_physical_layout(
            64,
            16,
            0,
            _buffer_and_sort_batches(
                16,
                store,
                queue,
            ),
        )
    )

    assert pa.Table.from_batches(batches).combine_chunks().to_batches() == [
        pa.RecordBatch.from_arrays(
            [
                pa.array(
                    [i if i % 2 == 0 else None for i in range(32)] + [None] * 32,
                    type=pa.uint64(),
                ),
                pa.array(range(64), type=pa.uint64()),
            ],
            names=["data", "_rowaddr"],
        )
    ]



#!/usr/bin/env python
# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Upload manifest and add columns for BLIP caption UDF to Geneva.

This script runs in the UDF's dependency environment (via uv run)
and uploads the manifest to the Geneva connection, then adds columns
to the specified table.

Usage:
    export GENEVA_TABLE_NAME=oxford_pets_shared_abc123
    cd e2e/oxford-pets/udfs/blip
    uv run python upload_manifest.py --bucket gs://bucket/path

Environment Variables:
    GENEVA_TABLE_NAME: Table name to add columns to (required)
    CAPTION_COL: Custom column name for caption (default: "caption")
"""

import logging
import os

from geneva.manifest.uploader_cli import main_upload_workflow
from blip_caption_udf import GenCaption
from manifest import create_manifest

logging.basicConfig(level=logging.INFO)
_LOG = logging.getLogger(__name__)


def get_columns() -> dict[str, callable]:
    """
    Get column definitions for this UDF package.

    Column names can be customized via environment variables:
    - CAPTION_COL: Custom name for caption column (default: "caption")

    Returns:
        Dictionary mapping column names to UDF functions
    """
    return {
        os.getenv("CAPTION_COL", "caption"): GenCaption(),
    }


if __name__ == "__main__":
    main_upload_workflow(
        get_columns_fn=get_columns,
        create_manifest_fn=create_manifest,
        description="Upload BLIP caption UDF manifest and add columns to table",
        default_manifest_name="blip-udfs-v1",
        logger=_LOG,
    )



import ast

import pytest
import ray
from kubernetes import config as kube_config
from kubernetes.client import ApiException
from ray.util.state.common import ActorState

from geneva.cluster import K8sConfigMethod
from geneva.runners.kuberay.client import KuberayClients, _wrap_api_methods
from geneva.runners.ray.kuberay import _ray_status, k8s_status

pytestmark = pytest.mark.ray

# ray.nodes() output
ray_nodes_json = """
[{'NodeID': 'b80f1624b03cd1af6714e2dc1d6f9f9eea14799ad52bde6057f44bf1',
  'Alive': True,
  'NodeManagerAddress': '172.17.78.133',
  'NodeManagerHostname': 'scorponok',
  'NodeManagerPort': 33027,
  'ObjectManagerPort': 45027,
  'ObjectStoreSocketName': '/tmp/ray/session_2025-06-17_15-05-30_021933_564383/sockets/plasma_store',
  'RayletSocketName': '/tmp/ray/session_2025-06-17_15-05-30_021933_564383/sockets/raylet',
  'MetricsExportPort': 64855,
  'NodeName': '172.17.78.133',
  'RuntimeEnvAgentPort': 53410,
  'DeathReason': 0,
  'DeathReasonMessage': '',
  'alive': True,
  'Resources': {'memory': 31255464756.0,
   'node:__internal_head__': 1.0,
   'GPU': 1.0,
   'object_store_memory': 13395199180.0,
   'node:172.17.78.133': 1.0,
   'CPU': 32.0,
   'accelerator_type:RTX': 1.0},
  'Labels': {'ray.io/node_id': 'b80f1624b03cd1af6714e2dc1d6f9f9eea14799ad52bde6057f44bf1'}}]
"""  # noqa: E501

# Sample dump from ray.util.state.list_actors
"""
[ActorState(actor_id='4d1e4fe662ba5e26aa06f46c01000000', class_name='ApplierActor',
    state='DEAD', job_id='01000000', name='file_size::images',
    node_id='e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95', pid=657551,
    ray_namespace='4490874b-64d4-4140-8195-c32247ccb2af', serialized_runtime_env=None,
    required_resources=None, death_cause=None, is_detached=None, placement_group_id=None,
    repr_name=None, num_restarts=None, num_restarts_due_to_lineage_reconstruction=None,
    call_site=None),
ActorState(actor_id='4e07c27f63f95807cafe1ea701000000', class_name='ProgressTracker',
    state='ALIVE', job_id='01000000', name='',
    node_id='e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95',
    pid=657553, ray_namespace='4490874b-64d4-4140-8195-c32247ccb2af',
    serialized_runtime_env=None, required_resources=None, death_cause=None,
    is_detached=None, placement_group_id=None, repr_name=None, num_restarts=None,
    num_restarts_due_to_lineage_reconstruction=None, call_site=None),
ActorState(actor_id='5e07c27f63f95807cafe1ea701000000', class_name='ProgressTracker',
    state='PENDING_CREATION', job_id='01000000', name='',
    node_id='e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95',
    pid=657553, ray_namespace='4490874b-64d4-4140-8195-c32247ccb2af',
    serialized_runtime_env=None, required_resources=None, death_cause=None,
    is_detached=None, placement_group_id=None, repr_name=None, num_restarts=None,
    num_restarts_due_to_lineage_reconstruction=None, call_site=None),
ActorState(actor_id='6e07c27f63f95807cafe1ea701000000', class_name='ProgressTracker',
    state='DEPENDENCIES_UNREADY', job_id='01000000', name='',
    node_id='e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95',
    pid=657553, ray_namespace='4490874b-64d4-4140-8195-c32247ccb2af',
    serialized_runtime_env=None, required_resources=None, death_cause=None,
    is_detached=None, placement_group_id=None, repr_name=None, num_restarts=None,
    num_restarts_due_to_lineage_reconstruction=None, call_site=None),
ActorState(actor_id='7e07c27f63f95807cafe1ea701000000', class_name='ProgressTracker',
    state='RESTARTING', job_id='01000000', name='',
    node_id='e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95',
    pid=657553, ray_namespace='4490874b-64d4-4140-8195-c32247ccb2af',
    serialized_runtime_env=None, required_resources=None, death_cause=None,
    is_detached=None, placement_group_id=None, repr_name=None, num_restarts=None,
    num_restarts_due_to_lineage_reconstruction=None, call_site=None),
ActorState(actor_id='8017a2c6bd91443240895bcd01000000', class_name='ProgressTracker',
    state='ALIVE', job_id='01000000', name='',
    node_id='e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95', pid=657555,
    ray_namespace='4490874b-64d4-4140-8195-c32247ccb2af', serialized_runtime_env=None,
    required_resources=None, death_cause=None, is_detached=None, placement_group_id=None,
    repr_name=None, num_restarts=None, num_restarts_due_to_lineage_reconstruction=None,
    call_site=None)]
"""  # noqa: E501


def make_actor_states() -> list[any]:
    return [
        ActorState(
            actor_id="4d1e4fe662ba5e26aa06f46c01000000",
            class_name="ApplierActor",
            state="DEAD",
            job_id="01000000",
            name="file_size::images",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657551,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
        ActorState(
            actor_id="4d1e4fe662ba5e26aa06f46c01000000",
            class_name="ApplierActor",
            state="ALIVE",
            job_id="01000000",
            name="file_size::images",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657552,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
        ActorState(
            actor_id="4e07c27f63f95807cafe1ea701000000",
            class_name="ProgressTracker",
            state="ALIVE",
            job_id="01000000",
            name="",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657553,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
        ActorState(
            actor_id="5e07c27f63f95807cafe1ea701000000",
            class_name="ApplierActor",
            state="PENDING_CREATION",
            job_id="01000000",
            name="",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657554,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
        ActorState(
            actor_id="6e07c27f63f95807cafe1ea701000000",
            class_name="ApplierActor",
            state="DEPENDENCIES_UNREADY",
            job_id="01000000",
            name="",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657553,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
        ActorState(
            actor_id="7e07c27f63f95807cafe1ea701000000",
            class_name="ApplierActor",
            state="RESTARTING",
            job_id="01000000",
            name="",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657553,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
        ActorState(
            actor_id="8017a2c6bd91443240895bcd01000000",
            class_name="ProgressTracker",
            state="ALIVE",
            job_id="01000000",
            name="",
            node_id="e227e86b72a0e9340c3eaec24cbdc2606830fc63b770583640205c95",
            pid=657555,
            ray_namespace="4490874b-64d4-4140-8195-c32247ccb2af",
        ),
    ]


def test_ray_status_with_real_actorstate(monkeypatch) -> None:
    # stub cluster_resources
    monkeypatch.setattr(ray, "cluster_resources", lambda: {"CPU": 32, "GPU": 1})
    nodes = ast.literal_eval(ray_nodes_json)
    monkeypatch.setattr(ray, "nodes", lambda: nodes)
    monkeypatch.setattr(
        "geneva.runners.ray.kuberay.list_actors",
        lambda **kwargs: make_actor_states(),
        raising=True,
    )

    # run
    status = _ray_status()

    # only one live node in our JSON
    assert status["cnt_ray_nodes"] == 1
    assert status["cnt_geneva_workers_active"] == 1
    assert status["cnt_geneva_workers_pending"] == 3

    # no error field
    assert "error" not in status


def test_ray_status_with_real_actorstate_no_ray(monkeypatch) -> None:
    # stub ray.nodes and list_actors to simulate no Ray cluster
    def raise_exception(ex) -> None:
        """lambdas cannot directly raise exceptions"""
        raise ex

    monkeypatch.setattr(
        ray, "nodes", lambda: raise_exception(Exception("not connected"))
    )
    monkeypatch.setattr(
        "geneva.runners.ray.kuberay.list_actors",
        lambda **kwargs: raise_exception(Exception("not connected")),
        raising=True,
    )

    # run
    status = _ray_status()

    # no  live nodes in our JSON
    assert status["cnt_ray_nodes"] == 0

    # error field
    assert "ray_nodes_error" in status


def _wrap_core_api(clients: KuberayClients) -> None:
    _wrap_api_methods(clients.core_api, lambda: clients.core_api, clients)  # type: ignore[arg-type]


def test_kuberay_clients_retries_401_and_refreshes(monkeypatch) -> None:
    # Avoid reading real kubeconfig in CI
    monkeypatch.setattr(kube_config, "load_kube_config", lambda: None)

    clients = KuberayClients(config_method=K8sConfigMethod.LOCAL)

    refresh_calls = {"n": 0}

    class _Always401:
        def delete_namespaced_config_map(self, name, namespace, **kwargs) -> dict:  # noqa: ANN001
            raise ApiException(status=401, reason="Unauthorized")

    class _AlwaysOK:
        def delete_namespaced_config_map(self, name, namespace, **kwargs) -> dict:  # noqa: ANN001
            return {"ok": True, "name": name, "ns": namespace}

    first_api = _Always401()
    second_api = _AlwaysOK()

    # Install first API and wrap methods
    clients.core_api = first_api  # type: ignore[assignment]
    _wrap_core_api(clients)

    def _refresh(self) -> None:  # type: ignore[no-redef]
        refresh_calls["n"] += 1
        # swap in a fresh API instance and re-wrap methods
        self.core_api = second_api  # type: ignore[assignment]
        _wrap_core_api(self)

    # monkeypatch refresh on the class so it can be undone cleanly
    monkeypatch.setattr(KuberayClients, "refresh", _refresh)

    # Call the wrapped method
    res = clients.core_api.delete_namespaced_config_map(  # type: ignore[attr-defined]
        name="cm-1", namespace="geneva"
    )

    assert res == {"ok": True, "name": "cm-1", "ns": "geneva"}
    assert refresh_calls["n"] >= 1


# Dummy classes to simulate Kubernetes pod objects
class DummyCondition:
    def __init__(self, _type: str, status: str) -> None:
        self.type = _type
        self.status = status


class DummyStatus:
    def __init__(self, phase: str, conditions: list[DummyCondition]) -> None:
        self.phase = phase
        self.conditions = conditions
        self.init_container_statuses = []
        self.container_statuses = []


class DummyMetadata:
    def __init__(self, name: str, labels: dict[str, str]) -> None:
        self.name = name
        self.labels = labels


class DummyPod:
    def __init__(
        self,
        name: str,
        labels: dict[str, str],
        phase: str,
        conditions: list[DummyCondition],
    ) -> None:
        self.metadata = DummyMetadata(name, labels)
        self.status = DummyStatus(phase, conditions)
        self.spec = None  # not used in the test


class DummyPodList:
    def __init__(self, items: list[DummyPod]) -> None:
        self.items = items


def test_k8s_status_filters_ray_pods_and_returns_details(monkeypatch) -> None:
    # Stub out kube_config.load_kube_config
    monkeypatch.setattr(kube_config, "load_kube_config", lambda: None)

    clients = KuberayClients(config_method=K8sConfigMethod.LOCAL)

    # Create a mix of pods, some with the ray label, some without
    pods = [
        DummyPod(
            name="ray-head-0",
            labels={"ray.io/node-type": "head"},
            phase="Running",
            conditions=[DummyCondition("Ready", "True")],
        ),
        DummyPod(
            name="other-pod",
            labels={},
            phase="Pending",
            conditions=[DummyCondition("Ready", "False")],
        ),
        DummyPod(
            name="ray-worker-1",
            labels={"ray.io/node-type": "worker"},
            phase="Pending",
            conditions=[DummyCondition("Ready", "False")],
        ),
    ]
    pod_list = DummyPodList(pods)

    # Stub out CoreV1Api.list_namespaced_pod
    class FakeCoreV1Api:
        def list_namespaced_pod(self, namespace) -> DummyPodList:
            assert namespace == "geneva"
            return pod_list

    monkeypatch.setattr(clients, "core_api", FakeCoreV1Api())

    # Call the function
    result = k8s_status(clients)

    # Verify that only the two ray pods are returned, in order
    assert len(result) == 2

    pod_statuses = [(r["name"], r["phase"], r["ready"]) for r in result]
    head_name, head_phase, head_ready = pod_statuses[0]
    assert head_name == "ray-head-0"
    assert head_phase == "Running"
    assert head_ready

    worker_name, worker_phase, worker_ready = pod_statuses[1]
    assert worker_name == "ray-worker-1"
    assert worker_phase == "Pending"
    assert not worker_ready


def test_k8s_status_returns_empty_when_no_ray_pods(monkeypatch) -> None:
    # Stub out kube_config.load_kube_config
    monkeypatch.setattr(kube_config, "load_kube_config", lambda: None)

    clients = KuberayClients(config_method=K8sConfigMethod.LOCAL)

    # Pod list with no ray.io/node-type labels
    pods = [
        DummyPod("app-1", {}, "Running", [DummyCondition("Ready", "True")]),
        DummyPod("app-2", {}, "Failed", [DummyCondition("Ready", "False")]),
    ]
    pod_list = DummyPodList(pods)

    class FakeCoreV1Api:
        def list_namespaced_pod(self, namespace) -> DummyPodList:
            assert namespace == "geneva"
            return pod_list

    monkeypatch.setattr(clients, "core_api", FakeCoreV1Api())

    result = k8s_status(clients)
    assert result == []  # no pods matching the ray label



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
from pathlib import Path

import lancedb
import pyarrow as pa
import pytest

from geneva import connect


def test_connect(tmp_path: Path) -> None:
    db = connect(tmp_path)

    # Use lancedb to verify the results are the same
    ldb = lancedb.connect(tmp_path)

    # Create a Table with integer columns
    tbl = pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
    db.create_table("table1", tbl)
    ldb_tbls = db.table_names()
    assert "table1" in ldb_tbls
    db.open_table("table1")

    db_tbls = db.table_names()
    assert db_tbls == ldb_tbls

    # Use lancedb to read the data back
    ldb_tbl = ldb.open_table("table1")
    assert ldb_tbl.to_arrow() == tbl
    db.drop_table("table1")


def test_connect_non_existent(tmp_path: Path) -> None:
    db = connect(tmp_path)
    assert db.table_names() == []

    with pytest.raises(ValueError, match=r".*was not found"):
        db.open_table("non_existent")



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import pytest

from geneva.cluster import K8sConfigMethod
from geneva.runners.ray.raycluster import RayCluster, _HeadGroupSpec, _WorkerGroupSpec


def test_service_account_does_not_exist(
    k8s_config_method: K8sConfigMethod,
    head_node_selector: dict,
    worker_node_selector: dict,
) -> None:
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        config_method=k8s_config_method,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        head_group=_HeadGroupSpec(node_selector=head_node_selector),  # type: ignore[call-arg]
        worker_groups=[
            _WorkerGroupSpec(  # type: ignore[call-arg]
                name="worker",
                min_replicas=1,
                service_account="does-not-exist",
                node_selector=worker_node_selector,
            )
        ],
    )
    with pytest.raises(
        ValueError, match="Service account does-not-exist does not exist"
    ):
        cluster._validate()


@pytest.mark.skip("requires RBAC permissions beyond what we require for Geneva users")
def test_service_account_not_enough_permission(
    k8s_temp_service_account: str,
    k8s_config_method: K8sConfigMethod,
    head_node_selector: dict,
    worker_node_selector: dict,
) -> None:
    # with strict access review, we should get an error about lack of permission
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        config_method=k8s_config_method,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        head_group=_HeadGroupSpec(node_selector=head_node_selector),  # type: ignore[call-arg]
        worker_groups=[
            _WorkerGroupSpec(  # type: ignore[call-arg]
                name="worker",
                min_replicas=1,
                service_account=k8s_temp_service_account,
                node_selector=worker_node_selector,
            )
        ],
        strict_access_review=True,
    )
    with pytest.raises(
        ValueError,
        match=f"Service account {k8s_temp_service_account} does not"
        " have the required permission:",
    ):
        cluster._validate()

    # with strict access review disabled, we should still get the same
    # error because we have permission to create local subject access
    # review and the service account does not have enough permission
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        config_method=k8s_config_method,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        head_group=_HeadGroupSpec(node_selector=head_node_selector),  # type: ignore[call-arg]
        worker_groups=[
            _WorkerGroupSpec(  # type: ignore[call-arg]
                name="worker",
                min_replicas=1,
                service_account=k8s_temp_service_account,
                node_selector=worker_node_selector,
            )
        ],
        strict_access_review=False,
    )
    with pytest.raises(
        ValueError,
        match=f"Service account {k8s_temp_service_account} does not"
        " have the required permission:",
    ):
        cluster._validate()


def test_service_account_is_valid(
    geneva_k8s_service_account: str,
    k8s_config_method: K8sConfigMethod,
    head_node_selector: dict,
    worker_node_selector: dict,
) -> None:
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        config_method=k8s_config_method,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        head_group=_HeadGroupSpec(node_selector=head_node_selector),  # type: ignore[call-arg]
        worker_groups=[
            _WorkerGroupSpec(  # type: ignore[call-arg]
                name="worker",
                min_replicas=1,
                service_account=geneva_k8s_service_account,
                node_selector=worker_node_selector,
            )
        ],
    )
    cluster._validate()



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from .builder import GenevaManifestBuilder
from .mgr import GenevaManifest, ManifestConfigManager

__all__ = ["GenevaManifest", "GenevaManifestBuilder", "ManifestConfigManager"]



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Pre-built embedding UDF helpers."""

import abc
import logging
import urllib.parse
from functools import cached_property
from typing import Any

import attrs
import pyarrow as pa

from geneva.transformer import UDF, udf

_LOG = logging.getLogger(__name__)

SENTENCE_TRANSFORMERS_FAMILY = "sentence-transformers"
DEFAULT_SENTENCE_TRANSFORMER_MODEL = "sentence-transformers/all-MiniLM-L6-v2"
DEFAULT_SENTENCE_TRANSFORMER_COLUMN = "text"


def _extract_string_inputs(
    batch: pa.RecordBatch, column: str
) -> tuple[list[str], list[int], list[Any]]:
    """Return valid string values and their row indices for ``column``."""

    index = batch.schema.get_field_index(column)
    if index == -1:
        raise ValueError(f"Column '{column}' not found in RecordBatch")

    field = batch.schema.field(index)
    if not pa.types.is_string(field.type) and not pa.types.is_large_string(field.type):
        raise TypeError(f"Column '{column}' must contain string data")

    values = batch.column(index).to_pylist()
    valid_indices: list[int] = []
    valid_texts: list[str] = []
    for idx, value in enumerate(values):
        if value is None:
            continue
        if not isinstance(value, str):
            raise TypeError(
                f"embedding UDF expects string inputs, received {type(value).__name__}."
            )
        valid_indices.append(idx)
        valid_texts.append(value)

    return valid_texts, valid_indices, values


def _resolve_device(num_gpus: float) -> str | None:
    if num_gpus <= 0:
        return None

    try:
        import torch
    except ImportError:
        _LOG.warning("torch not available; falling back to CPU for embeddings")
        return None

    if torch.cuda.is_available():
        _LOG.debug("CUDA is available; using GPU for embeddings")
        return "cuda"

    _LOG.debug("GPU requested but CUDA is not available; using CPU")
    return None


@attrs.define
class _EmbeddingModel(abc.ABC):
    """
    Base class interface for implementing pre-baked embedding models.
    All model families are required to only implement:
    * _build_model - load and return the model instance
    * _get_dimension - return the model's embedding dimension
    * embed - embed a RecordBatch and return a ListArray of float32 embeddings

    The base class handles lazy model loading, dimension caching, and output type.

    Parameters
    ----------
    model_name:
        The model being used for embedding. It can be a name of a model
        to be loaded from HuggingFace Hub, a local path or in case of
        API-based models, it can be an endpoint URL or model ID.
    column:
        Name of the column that will be embedded.
    normalize:
        Whether to L2-normalise the generated embeddings.
    num_gpus:
        GPU allocation requested for the UDF. Values ``>= 0``
        positive values will request CUDA based execution.
    """

    model_name: str = attrs.field(kw_only=True)
    column: str = attrs.field(kw_only=True)
    normalize: bool = attrs.field(kw_only=True)
    num_gpus: float = attrs.field(
        default=0.0, kw_only=True, validator=attrs.validators.ge(0)
    )

    _device: str | None = attrs.field(init=False, default=None)

    @abc.abstractmethod
    def _build_model(self) -> Any:
        """
        Return the model instance for the embedding backend. This
        method is called by UDF worker to lazily load the model.
        This can be used for one-time setup of the model instance, or
        client in case of API-based models (e.g. OpenAI, Gemini).
        """

    @abc.abstractmethod
    def _get_dimension(self) -> int:
        """
        Return the embedding dimension for the model.
        Embedding dimension must be a positive integer. It is the
        length of the embedding vector returned by the model for each input.
        """

    @abc.abstractmethod
    def embed(self, batch: pa.RecordBatch) -> pa.Array:
        """
        Embed ``batch`` and return a fixed-size list array of floats.
        It should handle missing inputs gracefully. For example::

            ["hello", None, "world"] -> [[0.1, 0.2], None, [0.3, 0.4]]
        """

    @cached_property
    def model(self) -> Any:
        """Lazily load and cache the model instance."""
        return self._build_model()

    @cached_property
    def dimension(self) -> int:
        """Lazily get and cache the model's embedding dimension."""
        dimension = self._get_dimension()
        if dimension <= 0:
            raise ValueError("embedding dimension must be a positive integer")
        return dimension

    def output_type(self) -> pa.DataType:
        return pa.list_(pa.float32(), self.dimension)

    def __getstate__(self) -> dict[str, Any]:
        """
        only serialize attributes excluding internal state or cached properties
        """
        return attrs.asdict(
            self,
            # include only the attributes that are part of __init__
            filter=lambda attribute, value: attribute.init,
        )

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__init__(**state)


@attrs.define
class _SentenceTransformersModel(_EmbeddingModel):
    trust_remote_code: bool = attrs.field(default=False, kw_only=True)

    def _build_model(self) -> Any:
        self._device = _resolve_device(self.num_gpus)
        model = self._load_model(self.model_name, self._device, self.trust_remote_code)
        return model

    def _get_dimension(self) -> int:
        dimension = self.model.get_sentence_embedding_dimension()
        return int(dimension)

    def embed(self, batch: pa.RecordBatch) -> pa.Array:
        model = self.model
        valid_texts, valid_indices, values = _extract_string_inputs(batch, self.column)

        outputs: list[list[float] | None] = [None] * len(values)
        if valid_texts:
            embeddings = model.encode(
                valid_texts,
                convert_to_numpy=True,
                normalize_embeddings=self.normalize,
            )
            vectors = embeddings.tolist()
            for idx, vector in zip(valid_indices, vectors, strict=False):
                outputs[idx] = vector

        return pa.array(outputs, type=self.output_type())

    @staticmethod
    def _load_model(
        model_name: str, device: str | None, trust_remote_code: bool
    ) -> Any:
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError as exc:
            raise ImportError(
                "sentence-transformers is required; install via "
                "`pip install sentence-transformers`"
            ) from exc

        return SentenceTransformer(
            model_name, device=device, trust_remote_code=trust_remote_code
        )


def _build_embedding_udf(
    model: _EmbeddingModel,
    udf_name: str,
    num_gpus: float,
    dimension: int | None = None,
) -> UDF:
    """
    Build an embedding UDF from a model.

    Parameters
    ----------
    model:
        The embedding model instance
    udf_name:
        Name for the UDF
    num_gpus:
        GPU allocation for the UDF
    dimension:
        Optional pre-specified embedding dimension. If None, will eagerly
        load the model to determine dimension. If provided, model loading
        is deferred until UDF execution.
    """
    if dimension is None:
        # Eager mode: load model now to get dimension
        dimension = model.dimension
        data_type = model.output_type()
    else:
        # Lazy mode: use provided dimension, defer model loading
        data_type = pa.list_(pa.float32(), dimension)

    @udf(
        name=udf_name,
        data_type=data_type,
        num_gpus=num_gpus,
    )
    class EmbeddingUDF:
        def __init__(self) -> None:
            self._model = model
            self.dimension = dimension

        def __call__(self, batch: pa.RecordBatch) -> pa.Array:
            return self._model.embed(batch)

    return EmbeddingUDF()  # type: ignore


# Supported pre-baked embedding UDFs


def sentence_transformer_udf(
    model: str = DEFAULT_SENTENCE_TRANSFORMER_MODEL,
    column: str = DEFAULT_SENTENCE_TRANSFORMER_COLUMN,
    normalize: bool = True,
    num_gpus: float = 0.0,
    trust_remote_code: bool = False,
    dimension: int | None = None,
) -> UDF:
    """
    Return a stateful sentence-transformers embedding UDF.

    Parameters
    ----------
    model:
        The model being used for embedding. by default, it uses
        ``sentence-transformers/all-MiniLM-L6-v2`` from HuggingFace Hub.
    column:
        Name of the column that will be embedded. By default, it uses ``text``.
    normalize:
        Whether to L2-normalise the generated embeddings. Defaults to ``True``.
    num_gpus:
        Fractional GPU allocation requested for the UDF. Values ``>= 0``
        Be default, keeps execution on CPU; positive values request CUDA.
    trust_remote_code:
        Whether to trust remote code when loading the model. Defaults to ``False``
        as recommended by sentence-transformers.
    dimension:
        Optional pre-specified embedding dimension. If None (default), will eagerly
        load the model to determine dimension. If provided, model loading is deferred
        until UDF execution. Use this for lazy loading when the model is not available
        at UDF definition time (e.g., in manifest upload scripts).

    Returns
    -------
    UDF
        A UDF instance that can be registered with a Geneva dataset.
    """

    embedding_model = _SentenceTransformersModel(
        model_name=model,
        column=column,
        normalize=normalize,
        num_gpus=num_gpus,
        trust_remote_code=trust_remote_code,
    )
    model_name_sanitized = urllib.parse.quote_plus(model)
    udf_name = f"{SENTENCE_TRANSFORMERS_FAMILY}:{model_name_sanitized}"
    return _build_embedding_udf(
        model=embedding_model,
        udf_name=udf_name,
        num_gpus=num_gpus,
        dimension=dimension,
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# a error logger for UDF jobs
import abc
from typing import TYPE_CHECKING

import attrs

if TYPE_CHECKING:
    from geneva.debug.error_store import ErrorRecord, ErrorStore


class ErrorLogger(abc.ABC):
    """Abstract interface for logging UDF execution errors"""

    @abc.abstractmethod
    def log_error(self, error: "ErrorRecord") -> None:
        """Log an error record

        Parameters
        ----------
        error : ErrorRecord
            The error record to log
        """
        ...

    @abc.abstractmethod
    def log_errors(self, errors: list["ErrorRecord"]) -> None:
        """Log multiple error records in bulk

        Parameters
        ----------
        errors : list[ErrorRecord]
            The error records to log
        """
        ...


class NoOpErrorLogger(ErrorLogger):
    """No-op error logger that discards all errors"""

    def log_error(self, error: "ErrorRecord") -> None:
        pass

    def log_errors(self, errors: list["ErrorRecord"]) -> None:
        pass


@attrs.define
class TableErrorLogger(ErrorLogger):
    """Error logger using ErrorStore (Lance table-based storage)"""

    error_store: "ErrorStore" = attrs.field()  # type: ignore[name-defined]

    def log_error(self, error: "ErrorRecord") -> None:
        """Log error record to error store table"""
        self.error_store.log_error(error)

    def log_errors(self, errors: list["ErrorRecord"]) -> None:
        """Log multiple error records in bulk"""
        self.error_store.log_errors(errors)



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import enum
import functools
import hashlib
import inspect
import logging
import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Optional

import attrs
import numpy
import pyarrow as pa
from attrs import validators as valid
from lance.blob import BlobFile

import geneva.cloudpickle as pickle

if TYPE_CHECKING:
    from geneva.debug.error_store import ErrorHandlingConfig

_LOG = logging.getLogger(__name__)

# special column name used to mark the rows that were not selected
# for backfilling.  This is used to avoid calling expensive UDFs
# on rows that are not selected.
BACKFILL_SELECTED = "__geneva_backfill_selected"


class UDFArgType(enum.Enum):
    """
    The type of arguments that the UDF expects.
    """

    # Scalar Batch
    SCALAR = 0
    # Array mode
    ARRAY = 1
    # Pass a pyarrow RecordBatch
    RECORD_BATCH = 2


@attrs.define
class UDF(Callable[[pa.RecordBatch], pa.Array]):  # type: ignore
    """User-defined function (UDF) to be applied to a Lance Table."""

    # The reference to the callable
    func: Callable = attrs.field()
    name: str = attrs.field(default="")
    cuda: Optional[bool] = attrs.field(default=False)
    num_cpus: Optional[float] = attrs.field(
        default=1.0,
        converter=lambda v: None if v is None else float(v),
        validator=valid.optional(valid.ge(0.0)),
        on_setattr=[attrs.setters.convert, attrs.setters.validate],
    )
    num_gpus: Optional[float] = attrs.field(
        default=None,
        converter=lambda v: None if v is None else float(v),
        validator=valid.optional(valid.ge(0.0)),
        on_setattr=[attrs.setters.convert, attrs.setters.validate],
    )
    memory: int | None = attrs.field(default=None)
    batch_size: int | None = attrs.field(default=None)

    # Error handling configuration
    error_handling: Optional["ErrorHandlingConfig"] = attrs.field(default=None)

    def _record_batch_input(self) -> bool:
        sig = inspect.signature(self.func)
        if len(sig.parameters) == 1:
            param = list(sig.parameters.values())[0]
            return param.annotation == pa.RecordBatch
        return False

    @property
    def arg_type(self) -> UDFArgType:
        if self._record_batch_input():
            return UDFArgType.RECORD_BATCH
        if _is_batched_func(self.func):
            return UDFArgType.ARRAY
        return UDFArgType.SCALAR

    input_columns: list[str] | None = attrs.field(default=None)

    data_type: pa.DataType = attrs.field(default=None)

    version: str = attrs.field(default="")

    checkpoint_key: str = attrs.field(default="")

    field_metadata: dict[str, str] = attrs.field(default={})

    def __attrs_post_init__(self) -> None:
        """
        Initialize UDF fields and normalize num_gpus after all fields are set:
          1) if cuda=True and num_gpus is None or 0.0 -> set to 1.0
          2) otherwise ignore cuda and just use num_gpus setting
        """
        # Set default name
        if not self.name:
            if inspect.isfunction(self.func):
                self.name = self.func.__name__
            elif isinstance(self.func, Callable):
                self.name = self.func.__class__.__name__
            else:
                raise ValueError(
                    f"func must be a function or a callable, got {self.func}"
                )

        # Set default input_columns
        if self.input_columns is None:
            sig = inspect.signature(self.func)
            params = list(sig.parameters.keys())
            if self._record_batch_input():
                self.input_columns = None
            else:
                self.input_columns = params

        # Validate input_columns
        if self.arg_type == UDFArgType.RECORD_BATCH:
            if self.input_columns is not None:
                raise ValueError(
                    "RecordBatch input UDF must not declare any input columns. "
                    "RecordBatch UDFs receive the entire batch and should not "
                    "specify input_columns. Consider using a stateful RecordBatch "
                    "UDF and parameterize it or use UDF with Array inputs."
                )
        else:
            if self.input_columns is None:
                raise ValueError("Array and Scalar input UDF must declare input column")

        # Set default data_type
        if self.data_type is None:
            if self.arg_type != UDFArgType.SCALAR:
                raise ValueError(
                    "batched UDFs do not support data_type inference yet,"
                    " please specify data_type",
                )
            self.data_type = _infer_func_arrow_type(self.func, None)  # type: ignore[arg-type]

        # Validate data_type
        if self.data_type is None:
            raise ValueError("data_type must be set")
        if not isinstance(self.data_type, pa.DataType):
            raise ValueError(
                f"data_type must be a pyarrow.DataType, got {self.data_type}"
            )

        # Set default version
        if not self.version:
            hasher = hashlib.md5()
            hasher.update(pickle.dumps(self.func))
            self.version = hasher.hexdigest()

        # Set default checkpoint_key
        if not self.checkpoint_key:
            self.checkpoint_key = f"{self.name}:{self.version}"

        # Handle cuda/num_gpus normalization
        if self.cuda:
            warnings.warn(
                "The 'cuda' flag is deprecated. Please set 'num_gpus' explicitly "
                "(0.0 for CPU, >=1.0 for GPU).",
                DeprecationWarning,
                stacklevel=2,
            )

        if self.num_gpus is None:
            self.num_gpus = 1.0 if self.cuda is True else 0.0
        # otherwise fall back to user specified num_gpus

    def __repr__(self) -> str:
        """Custom repr that safely handles missing attributes during unpickling.

        This is necessary because attrs-generated __repr__ can fail when called
        during exception handling in Ray if the object hasn't been fully unpickled yet.
        """
        try:
            # Try to get all attrs fields safely
            field_strs = []
            for field in attrs.fields(self.__class__):
                # Check if attribute exists first before accessing it
                if hasattr(self, field.name):
                    value = getattr(self, field.name)
                    field_strs.append(f"{field.name}={value!r}")
                else:
                    field_strs.append(f"{field.name}=<not set>")

            return f"{self.__class__.__qualname__}({', '.join(field_strs)})"
        except Exception:
            # Fallback if even that fails
            return f"<{self.__class__.__name__} (repr failed)>"

    def _scalar_func_record_batch_call(self, record_batch: pa.RecordBatch) -> pa.Array:
        """
        We use this when the UDF uses single call like
        `func(x_int, y_string, ...) -> type`

        this function automatically dispatches rows to the func and returns `pa.Array`
        """

        # this let's us avoid having to allocate a list in python
        # to hold the results. PA will allocate once for us
        def _iter():  # noqa: ANN202
            batches = (
                record_batch.to_pylist()
                if isinstance(record_batch, pa.RecordBatch)
                else record_batch
            )
            for item in batches:
                # we know inputs_columns is not none here
                if BACKFILL_SELECTED not in item or item.get(BACKFILL_SELECTED):
                    # we know input_columns is not none here
                    args = [item[col] for col in self.input_columns]  # type: ignore
                    yield self.func(*args)
                else:
                    # item was not selected, so do not compute
                    yield None

        arr = pa.array(
            _iter(),
            type=self.data_type,
        )
        # this should always by an Array, never should we get a ChunkedArray back here
        assert isinstance(arr, pa.Array)
        return arr

    def _input_columns_validator(self, attribute, value) -> None:
        """Validate input_columns attribute for attrs compatibility."""
        if self.arg_type == UDFArgType.RECORD_BATCH:
            if value is not None:
                raise ValueError(
                    "RecordBatch input UDF must not declare any input columns. "
                    "RecordBatch UDFs receive the entire batch and should not "
                    "specify input_columns."
                )
        else:
            if value is None:
                raise ValueError("Array and Scalar input UDF must declare input column")

    def validate_against_schema(
        self, table_schema: pa.Schema, input_columns: list[str] | None = None
    ) -> None:
        """
        Validate UDF against table schema.

        This is the primary validation method that should be called before executing
        a UDF. It performs comprehensive validation including:

        1. **Column Existence**: Verifies all input columns exist in the table schema
        2. **Type Compatibility**: Checks that column types match UDF type annotations
           (if present)
        3. **RecordBatch Constraints**: Ensures RecordBatch UDFs don't have
           input_columns defined

        The validation happens at two points in the UDF lifecycle:
        - At `add_columns()` time when defining the column
        - At `backfill()` time when executing (if input_columns are overridden)

        Parameters
        ----------
        table_schema: pa.Schema
            The schema of the table being processed
        input_columns: list[str] | None
            The input column names to validate. If None, uses self.input_columns.

        Raises
        ------
        ValueError: If validation fails for any of the following reasons:
            - Input columns don't exist in table schema
            - Type mismatch between table and UDF expectations
            - RecordBatch UDF has input_columns defined
            - Array/Scalar UDF has no input_columns defined

        Warns
        -----
        UserWarning: If type validation is skipped due to:
            - UDF has no type annotations
            - Type annotation can't be mapped to PyArrow types

        Examples
        --------
        >>> @udf(data_type=pa.int32())
        ... def my_udf(a: int) -> int:
        ...     return a * 2
        >>> my_udf.validate_against_schema(table.schema)  # Validates column 'a' exists
        """

        # Determine which columns to validate
        cols_to_validate = (
            input_columns if input_columns is not None else self.input_columns
        )

        # Check RecordBatch UDFs
        if self.arg_type == UDFArgType.RECORD_BATCH:
            # Error if input_columns are specified for RecordBatch UDFs
            if cols_to_validate is not None:
                raise ValueError(
                    f"UDF '{self.name}' is a RecordBatch UDF but has input_columns "
                    f"{cols_to_validate} specified. RecordBatch UDFs receive the "
                    f"entire batch and should not declare input_columns. "
                    f"Remove the input_columns parameter."
                )
            # RecordBatch UDFs don't need column validation
            return

        # For Array and Scalar UDFs, input_columns must be defined
        if cols_to_validate is None:
            arg_type_name = self.arg_type.name if self.arg_type else "UNKNOWN"
            raise ValueError(
                f"UDF '{self.name}' (type: {arg_type_name}) has no input_columns "
                f"defined. Array and Scalar UDFs must specify input columns either "
                f"through function parameter names or the input_columns parameter."
            )

        # Validate all input columns exist in table schema
        missing_columns = [
            col for col in cols_to_validate if col not in table_schema.names
        ]

        if missing_columns:
            raise ValueError(
                f"UDF '{self.name}' expects input columns {missing_columns} which are "
                f"not found in table schema. Available columns: {table_schema.names}. "
                f"Check your UDF's function parameter names or input_columns parameter."
            )

        # Validate type compatibility for each input column
        self._validate_column_types(table_schema, cols_to_validate)

    def _validate_column_types(
        self, table_schema: pa.Schema, input_columns: list[str]
    ) -> None:
        """
        Validate type compatibility between table schema and UDF expectations.

        This method checks if the table column types match the UDF's type annotations.
        If no type annotations are present or types can't be mapped, validation is
        skipped with a warning.

        Parameters
        ----------
        table_schema: pa.Schema
            The schema of the table being processed
        input_columns: list[str]
            The input column names to validate types for

        Raises
        ------
        ValueError: If there's a type mismatch between table schema and UDF expectations

        Warns
        -----
        UserWarning: If type validation is skipped due to missing annotations or
            unmappable types
        """
        import warnings

        # Get type annotations from the UDF function
        annotations = _get_annotations(self.func)

        if not annotations:
            # No type annotations found - warn user
            warnings.warn(
                f"UDF '{self.name}' has no type annotations. Type validation will be "
                f"skipped. Consider adding type hints to your UDF function parameters "
                f"for better error detection.",
                UserWarning,
                stacklevel=4,
            )
            return

        # For each input column, validate type if annotation exists
        for col_name in input_columns:
            # Get the actual type from table schema
            table_field = table_schema.field(col_name)
            table_type = table_field.type

            # Get expected type from UDF signature if available
            if col_name in annotations:
                expected_type = annotations[col_name]

                # Try to map expected type to PyArrow type for comparison
                try:
                    expected_pa_type = self._python_type_to_arrow_type(expected_type)

                    # Check if types are compatible
                    if not self._types_compatible(table_type, expected_pa_type):
                        raise ValueError(
                            f"Type mismatch for column '{col_name}' in UDF "
                            f"'{self.name}': table has type {table_type}, but UDF "
                            f"expects {expected_pa_type} (from annotation "
                            f"{expected_type}). This will likely cause serialization "
                            f"or conversion errors during execution."
                        )
                except (ValueError, KeyError):
                    # If we can't map the type, skip validation with warning
                    warnings.warn(
                        f"Could not validate type for column '{col_name}' in UDF "
                        f"'{self.name}' with annotation {expected_type}. Type "
                        f"validation skipped for this column.",
                        UserWarning,
                        stacklevel=4,
                    )

    def _python_type_to_arrow_type(self, python_type) -> pa.DataType:
        """
        Convert Python type annotation to PyArrow type.

        Raises ValueError if type cannot be mapped.
        """
        # Handle PyArrow types directly
        if isinstance(python_type, pa.DataType):
            return python_type

        # Handle pa.Array annotation (for batched UDFs)
        if python_type == pa.Array:
            # Can't determine specific array type, so return None to skip validation
            raise ValueError("Cannot validate generic pa.Array type")

        # Map Python/numpy types to PyArrow types
        type_map = {
            bool: pa.bool_(),
            bytes: pa.binary(),
            float: pa.float32(),
            int: pa.int64(),
            str: pa.string(),
            numpy.bool_: pa.bool_(),
            numpy.uint8: pa.uint8(),
            numpy.uint16: pa.uint16(),
            numpy.uint32: pa.uint32(),
            numpy.uint64: pa.uint64(),
            numpy.int8: pa.int8(),
            numpy.int16: pa.int16(),
            numpy.int32: pa.int32(),
            numpy.int64: pa.int64(),
            numpy.float16: pa.float16(),
            numpy.float32: pa.float32(),
            numpy.float64: pa.float64(),
            numpy.str_: pa.string(),
        }

        if python_type in type_map:
            return type_map[python_type]

        raise ValueError(f"Cannot map Python type {python_type} to PyArrow type")

    def _types_compatible(self, actual: pa.DataType, expected: pa.DataType) -> bool:
        """
        Check if actual type is compatible with expected type.

        This is more permissive than exact equality, allowing for:
        - Exact matches
        - Nullable vs non-nullable variants
        """
        # Exact match
        if actual == expected:
            return True

        # Check base types match (ignoring nullability, precision differences)
        # For numeric types, check if they're in the same family
        if pa.types.is_integer(actual) and pa.types.is_integer(expected):
            # Allow integer types if bit width and signedness match
            return actual.bit_width == expected.bit_width and (
                (
                    pa.types.is_signed_integer(actual)
                    and pa.types.is_signed_integer(expected)
                )
                or (
                    pa.types.is_unsigned_integer(actual)
                    and pa.types.is_unsigned_integer(expected)
                )
            )

        if pa.types.is_floating(actual) and pa.types.is_floating(expected):
            # Require exact match for floating point types (float32 vs float64 matters!)
            return actual.bit_width == expected.bit_width

        # For other types, require exact match
        return False

    def __call__(self, *args, use_applier: bool = False, **kwargs) -> pa.Array:
        # dispatch coming from Applier or user calling with a `RecordBatch`
        if use_applier or (len(args) == 1 and isinstance(args[0], pa.RecordBatch)):
            record_batch = args[0]
            match self.arg_type:
                case UDFArgType.SCALAR:
                    return self._scalar_func_record_batch_call(record_batch)
                case UDFArgType.ARRAY:
                    # Validate columns exist before accessing them
                    try:
                        arrs = [record_batch[col] for col in self.input_columns]  # type:ignore
                    except KeyError as e:
                        raise KeyError(
                            f"UDF '{self.name}' failed: column {e} not found in "
                            f"RecordBatch. Available columns: "
                            f"{record_batch.schema.names}. UDF expects "
                            f"input_columns: {self.input_columns}."
                        ) from e
                    return self.func(*arrs)
                case UDFArgType.RECORD_BATCH:
                    if isinstance(record_batch, pa.RecordBatch):
                        return self.func(record_batch)
                    # a list of dicts with BlobFiles that need to de-ref'ed
                    assert isinstance(record_batch, list)
                    rb_list = []
                    for row in record_batch:
                        new_row = {}
                        for k, v in row.items():
                            if isinstance(v, BlobFile):
                                # read the blob file into memory
                                new_row[k] = v.readall()
                                continue
                            new_row[k] = v
                        rb_list.append(new_row)

                    rb = pa.RecordBatch.from_pylist(rb_list)

                    return self.func(rb)
        # dispatch is trying to access the function's original pattern
        return self.func(*args, **kwargs)


def udf(
    func: Callable | None = None,
    *,
    data_type: pa.DataType | None = None,
    version: str | None = None,
    cuda: bool = False,  # deprecated
    field_metadata: dict[str, str] | None = None,
    input_columns: list[str] | None = None,
    num_cpus: int | float | None = None,
    num_gpus: int | float | None = None,
    error_handling: Optional["ErrorHandlingConfig"] = None,
    **kwargs,
) -> UDF | functools.partial:
    """Decorator of a User Defined Function ([UDF][geneva.transformer.UDF]).

    Parameters
    ----------
    func: Callable
        The callable to be decorated. If None, returns a partial function.
    data_type: pa.DataType, optional
        The data type of the output PyArrow Array from the UDF.
        If None, it will be inferred from the function signature.
    version: str, optional
        A version string to manage the changes of function.
        If not provided, it will use the hash of the serialized function.
    cuda: bool, optional, Deprecated
        If true, load CUDA optimized kernels.  Equvalent to num_gpus=1
    field_metadata: dict[str, str], optional
        A dictionary of metadata to be attached to the output `pyarrow.Field`.
    input_columns: list[str], optional
        A list of input column names for the UDF. If not provided, it will be
        inferred from the function signature. Or scan all columns.
    num_cpus: int, float, optional
        The (fraction) number of CPUs to acquire to run the job.
    num_gpus: int, float, optional
        The (fraction) number of GPUs to acquire to run the job.  Default 0.
    error_handling: ErrorHandlingConfig, optional
        Configuration for error handling and retry behavior.
    """
    if inspect.isclass(func):

        @functools.wraps(func)
        def _wrapper(*args, **kwargs) -> UDF | functools.partial:
            callable_obj = func(*args, **kwargs)
            return udf(
                callable_obj,
                cuda=cuda,
                data_type=data_type,
                version=version,
                field_metadata=field_metadata,
                input_columns=input_columns,
                num_cpus=num_cpus,
                num_gpus=num_gpus,
                error_handling=error_handling,
            )

        return _wrapper  # type: ignore

    if func is None:
        return functools.partial(
            udf,
            cuda=cuda,
            data_type=data_type,
            version=version,
            field_metadata=field_metadata,
            input_columns=input_columns,
            num_cpus=num_cpus,
            num_gpus=num_gpus,
            error_handling=error_handling,
            **kwargs,
        )

    # we depend on default behavior of attrs to infer the output schema
    def _include_if_not_none(name, value) -> dict[str, Any]:
        if value is not None:
            return {name: value}
        return {}

    args = {
        "func": func,
        "cuda": cuda,
        **_include_if_not_none("data_type", data_type),
        **_include_if_not_none("version", version),
        **_include_if_not_none("field_metadata", field_metadata),
        **_include_if_not_none("input_columns", input_columns),
        **_include_if_not_none("num_cpus", num_cpus),
        **_include_if_not_none("num_gpus", num_gpus),
        **_include_if_not_none("error_handling", error_handling),
    }
    # can't use functools.update_wrapper because attrs makes certain assumptions
    # and attributes read-only. We will figure out docs and stuff later
    return UDF(**args)


def _get_annotations(func: Callable) -> dict[str, type]:
    if inspect.isfunction(func):
        return inspect.get_annotations(func)
    elif isinstance(func, Callable):
        return inspect.get_annotations(func.__call__)
    raise ValueError(f"func must be a function or a callable, got {func}")


def _is_batched_func(func: Callable) -> bool:
    annotations = _get_annotations(func)
    if "return" not in annotations:
        return False

    ret_type = annotations["return"]
    if ret_type != pa.Array and not isinstance(ret_type, pa.DataType):
        return False

    input_keys = list(annotations.keys() - {"return"})
    if len(input_keys) == 1:
        return all(
            annotations[input_key] in [pa.RecordBatch, pa.Array]
            for input_key in input_keys
        )

    if any(annotations[input_key] == pa.RecordBatch for input_key in input_keys):
        raise ValueError(
            "UDF can not have multiple parameters with 'pa.RecordBatch' type"
        )
    return all(annotations[input_key] in [pa.Array] for input_key in input_keys)


# Build numpy type mapping - numpy.bool deprecated in 1.x, reintroduced in 2.x
_NUMPY_TYPE_MAP = {
    bool: pa.bool_(),
    bytes: pa.binary(),
    float: pa.float32(),
    int: pa.int64(),
    str: pa.string(),
    numpy.bool_: pa.bool_(),
    numpy.uint8: pa.uint8(),
    numpy.uint16: pa.uint16(),
    numpy.uint32: pa.uint32(),
    numpy.uint64: pa.uint64(),
    numpy.int8: pa.int8(),
    numpy.int16: pa.int16(),
    numpy.int32: pa.int32(),
    numpy.int64: pa.int64(),
    numpy.float16: pa.float16(),
    numpy.float32: pa.float32(),
    numpy.float64: pa.float64(),
    numpy.str_: pa.string(),
}

# Add numpy.bool if available (numpy 2.x)
# In numpy 2.x, numpy.bool is a proper type, not a deprecated alias
if hasattr(numpy, "bool") and isinstance(numpy.bool, type):
    _NUMPY_TYPE_MAP[numpy.bool] = pa.bool_()


def _infer_func_arrow_type(func: Callable, input_schema: pa.Schema) -> pa.DataType:
    """Infer the output schema of a UDF

    currently independent of the input schema, in the future we may want to
    infer the output schema based on the input schema, or the UDF itself could
    request the input schema to be passed in.
    """
    if isinstance(func, UDF):
        return func.data_type

    annotations = _get_annotations(func)
    if "return" not in annotations:
        raise ValueError(f"UDF {func} does not have a return type annotation")

    data_type = annotations["return"]
    # do dispatch to handle different types of output types
    # e.g. pydantic -> pyarrow type inference
    if isinstance(data_type, pa.DataType):
        return data_type

    if t := _NUMPY_TYPE_MAP.get(data_type):
        return t

    raise ValueError(f"UDF {func} has an invalid return type annotation {data_type}")



#!/usr/bin/env python
# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Upload manifest and add columns for simple OpenVid UDFs to Geneva.

This script runs in the UDF's dependency environment (via uv run)
and uploads the manifest to the Geneva connection, then adds columns
to the specified table.

Usage:
    export GENEVA_TABLE_NAME=openvid_shared_abc123
    cd e2e/openvid/udfs/simple
    uv run python upload_manifest.py --bucket gs://bucket/path

Environment Variables:
    GENEVA_TABLE_NAME: Table name to add columns to (required)
    HAS_FILE_COL: Custom column name for has_file (default: "has_file")
"""

import logging
import os

from geneva.manifest.uploader_cli import main_upload_workflow
from manifest import create_manifest
from simple_udfs import HasFile

logging.basicConfig(level=logging.INFO)
_LOG = logging.getLogger(__name__)


def get_columns() -> dict[str, callable]:
    """
    Get column definitions for this UDF package.

    Column names can be customized via environment variables:
    - HAS_FILE_COL: Custom name for has_file column (default: "has_file")

    Returns:
        Dictionary mapping column names to UDF functions
    """
    return {
        os.getenv("HAS_FILE_COL", "has_file"): HasFile(),
    }


if __name__ == "__main__":
    main_upload_workflow(
        get_columns_fn=get_columns,
        create_manifest_fn=create_manifest,
        description="Upload simple OpenVid UDFs manifest and add columns to table",
        default_manifest_name="openvid-simple-udfs-v1",
        logger=_LOG,
    )



# ruff: noqa: E402
# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# lance dataset distributed transform job checkpointing + UDF utils

import base64
import fcntl
import json
import logging
import os
import site
import tempfile
import threading
import zipfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

_LOG = logging.getLogger(__name__)

_extract_lock = threading.Lock()


def _download_and_extract(args: dict) -> None:
    # Import pyarrow lazily to avoid issues when pyarrow isn't available yet
    # (e.g., when it's part of the manifest being loaded via GENEVA_ZIPS)
    import pyarrow.fs as fs

    src_path, download_path, output_dir = (
        args["src_path"],
        args["download_path"],
        args["output_dir"],
    )
    handle, path = fs.FileSystem.from_uri(src_path)
    handle: fs.FileSystem = handle
    path: str = path

    with (
        handle.open_input_file(path) as f,
        open(download_path, "wb") as out,
    ):
        chunk_size = 1024 * 1024 * 8  # 8MiB chunks
        while data := f.read(chunk_size):
            out.write(data)

    with (
        zipfile.ZipFile(download_path) as z,
        _extract_lock,  # ensure only one thread extracts at a time
    ):
        z.extractall(output_dir)
        _LOG.info("extracted workspace to %s", output_dir)


# MAGIC: if GENEVA_ZIPS is set, we will extract the zips and add them as site-packages
# this is how we acheive "import geneva" == importing workspace from client
#
# NOTE: think of this like booting up a computer. At this point we do not have any
# dependencies installed, so this logic needs to have minimal dependency surface.
# We avoid importing anything from geneva and do everything in the stdlib
if "GENEVA_ZIPS" in os.environ:
    import fcntl

    with (
        open("/tmp/.geneva_zip_setup", "w") as file,  # noqa: S108
        ThreadPoolExecutor(max_workers=8) as executor,
    ):
        # use fcntl to lock the file so we don't have multiple processes
        # trying to extract at the same time and blow up the disk space
        fcntl.lockf(file, fcntl.LOCK_EX)

        payload = json.loads(base64.b64decode(os.environ["GENEVA_ZIPS"]))
        zips = payload.get("zips", [])

        for parts in zips:
            if not len(parts):
                # got an empty list, skip
                continue

            _LOG.info("Setting up geneva workspace from zips %s", parts)
            file_name = parts[0].split("/")[-1]
            name = file_name.split(".")[0]
            output_dir = Path(tempfile.gettempdir()) / name
            if output_dir.exists():
                _LOG.info("workspace already extracted to %s", output_dir)
            else:
                # force collect to surface errors
                list(
                    executor.map(
                        _download_and_extract,
                        (
                            {
                                "src_path": z,
                                "download_path": Path(tempfile.gettempdir())
                                / z.split("/")[-1],
                                "output_dir": output_dir,
                            }
                            for z in parts
                        ),
                    )
                )

            site.addsitedir(output_dir.as_posix())
            _LOG.info("added %s to sys.path", output_dir)

        fcntl.lockf(file, fcntl.LOCK_UN)


from geneva._context import get_current_context
from geneva.apply import CheckpointingApplier, ReadTask, ScanTask
from geneva.checkpoint import (
    CheckpointStore,
    InMemoryCheckpointStore,
)
from geneva.db import connect
from geneva.transformer import udf

__all__ = [
    "CheckpointStore",
    "connect",
    "InMemoryCheckpointStore",
    "CheckpointingApplier",
    "ReadTask",
    "ScanTask",
    "udf",
    "get_current_context",
]

version = "0.7.0"

__version__ = version



from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

import pytest

from geneva.cluster.builder import GenevaClusterBuilder
from geneva.manifest import GenevaManifestBuilder


class DummyClusterDef:
    def __init__(self, cluster_type: Any, use_portforwarding: bool = True) -> None:
        self.cluster_type = cluster_type
        self._use_portforwarding = use_portforwarding

    def as_dict(self) -> dict[str, Any]:
        return {"kuberay": {"use_portforwarding": self._use_portforwarding}}

    def to_ray_cluster(self) -> Any:
        # Return a dummy RayCluster object to capture on_exit attribute
        return type("RC", (), {})()


class DummyManifestDef:
    @property
    def head_image(self) -> str | None:
        return None

    @property
    def worker_image(self) -> str | None:
        return None


@pytest.fixture
def patch_dependencies(monkeypatch: Any) -> dict[str, Any]:
    # Patch ray_cluster and prevent real manager initialization
    import geneva.cluster.mgr as cluster_mgr_mod
    import geneva.manifest.mgr as manifest_mgr_mod
    import geneva.runners.ray._mgr as ray_mgr_mod

    called = {}

    @contextmanager
    def dummy_ray_cluster(**kwargs: Any) -> Generator[str, None, None]:
        called.update(kwargs)
        yield "ctx"

    # Patch ray_cluster in the module where it's defined
    monkeypatch.setattr(ray_mgr_mod, "ray_cluster", dummy_ray_cluster)
    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "__init__", lambda self, db: None
    )
    monkeypatch.setattr(
        manifest_mgr_mod.ManifestConfigManager, "__init__", lambda self, db: None
    )
    return called


def test_ray_cluster_local(monkeypatch: Any) -> None:
    # Import modules inside function to avoid import errors at module level
    import geneva.packager.autodetect as autodetect_mod
    import geneva.runners.ray._mgr as ray_mgr_mod

    # local=True should skip upload and call init_ray with empty zips
    called = {}

    @contextmanager
    def dummy_init_ray(*args: Any, **kwargs: Any) -> Generator[None, None, None]:
        called["init"] = kwargs.copy()
        yield

    @contextmanager
    def dummy_upload_local_env(
        *args: Any, **kwargs: Any
    ) -> Generator[list[list[str]], None, None]:
        called["upload"] = True
        yield [["dummy.zip"]]

    # Mock the functions in ray_mgr_mod where they are used
    monkeypatch.setattr(ray_mgr_mod, "init_ray", dummy_init_ray)
    monkeypatch.setattr(ray_mgr_mod, "upload_local_env", dummy_upload_local_env)
    monkeypatch.setattr(autodetect_mod, "upload_local_env", dummy_upload_local_env)

    # Also mock ray.init and ray.shutdown to prevent actual Ray operations
    monkeypatch.setattr("ray.init", lambda *args, **kwargs: None)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Mock base64 and json to prevent encoding operations
    monkeypatch.setattr("base64.b64encode", lambda x: b"mocked_base64")
    monkeypatch.setattr("json.dumps", lambda x: "mocked_json")

    # invoke local ray cluster
    with ray_mgr_mod.ray_cluster(local=True):
        pass
    # upload_local_env should not be called, init_ray zips should be empty
    assert "upload" not in called
    assert called["init"]["local"] is True
    assert called["init"]["zips"] == []


def test_ray_cluster_new_cluster(monkeypatch: Any) -> None:
    # Import modules inside function to avoid import errors at module level
    import geneva.packager.autodetect as autodetect_mod
    import geneva.runners.ray._mgr as ray_mgr_mod

    # non-local cluster: stub RayCluster and PortForward to capture addr and zips
    called = {}
    dummy_zips = [["a.zip", "b.zip"]]

    @contextmanager
    def dummy_upload_local_env(
        *args: Any, **kwargs: Any
    ) -> Generator[list[list[str]], None, None]:
        called["upload_opts"] = kwargs.copy()
        yield dummy_zips

    @contextmanager
    def dummy_init_ray(*args: Any, **kwargs: Any) -> Generator[None, None, None]:
        called["init"] = kwargs.copy()
        yield

    class DummyRC:
        def __init__(self, **kwargs: Any) -> None:
            called["rc_opts"] = kwargs.copy()
            self.ray_init_kwargs = None  # Mock the ray_init_kwargs attribute
            self.manifest = None  # Mock the manifest attribute

        def __enter__(self) -> str:
            return "10.0.0.1"

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            pass

    @contextmanager
    def dummy_pf(cluster: Any) -> Generator[Any, None, None]:
        class PF:
            local_port = "9999"

        yield PF()

    # Mock the functions and classes in ray_mgr_mod where they are used
    monkeypatch.setattr(ray_mgr_mod, "upload_local_env", dummy_upload_local_env)
    monkeypatch.setattr(ray_mgr_mod, "init_ray", dummy_init_ray)
    monkeypatch.setattr(ray_mgr_mod, "RayCluster", DummyRC)
    monkeypatch.setattr(
        ray_mgr_mod.PortForward, "to_head_node", classmethod(lambda cls, c: dummy_pf(c))
    )
    monkeypatch.setattr(
        ray_mgr_mod.PortForward, "to_ui", classmethod(lambda cls, c: dummy_pf(c))
    )
    monkeypatch.setattr(autodetect_mod, "upload_local_env", dummy_upload_local_env)

    # Also mock ray.init and ray.shutdown to prevent actual Ray operations
    monkeypatch.setattr("ray.init", lambda *args, **kwargs: None)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Mock base64 and json to prevent encoding operations
    monkeypatch.setattr("base64.b64encode", lambda x: b"mocked_base64")
    monkeypatch.setattr("json.dumps", lambda x: "mocked_json")

    with ray_mgr_mod.ray_cluster():
        pass
    # upload_local_env called with default flags
    assert "upload_opts" in called
    # init_ray should be called with computed addr and zips
    assert called["init"]["addr"] == "ray://localhost:9999"
    assert called["init"]["zips"] == dummy_zips


def test_local_ray_no_manifest(
    tmp_path: Path, patch_dependencies: dict[str, Any]
) -> None:
    from geneva import connect
    from geneva.cluster import GenevaClusterType

    conn = connect(tmp_path)
    called = patch_dependencies
    # LOCAL_RAY should invoke ray_cluster with local=True
    with conn.context(cluster_type=GenevaClusterType.LOCAL_RAY) as ctx:
        assert ctx == "ctx"
    assert called == {"local": True, "log_to_driver": True, "logging_level": 20}


def test_local_ray_manifest_error(
    tmp_path: Path, patch_dependencies: dict[str, Any]
) -> None:
    from geneva import connect
    from geneva.cluster import GenevaClusterType

    conn = connect(tmp_path)
    with (
        pytest.raises(ValueError, match="custom manifest not supported with LOCAL_RAY"),
        conn.context(cluster_type=GenevaClusterType.LOCAL_RAY, manifest="m"),
    ):
        pass


def test_cluster_not_found_error(
    tmp_path: Path, patch_dependencies: dict[str, Any], monkeypatch: Any
) -> None:
    import geneva.cluster.mgr as cluster_mgr_mod
    from geneva import connect

    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "load", lambda self, name: None
    )
    conn = connect(tmp_path)
    with (
        pytest.raises(Exception, match="cluster definition 'abc' not found"),
        conn.context(cluster="abc"),
    ):
        pass


def test_unsupported_cluster_type_error(
    tmp_path: Path, patch_dependencies: dict[str, Any], monkeypatch: Any
) -> None:
    import geneva.cluster.mgr as cluster_mgr_mod
    from geneva import connect
    from geneva.cluster import GenevaClusterType

    dummy = DummyClusterDef(GenevaClusterType.LOCAL_RAY)
    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "load", lambda self, name: dummy
    )
    conn = connect(tmp_path)
    with (
        pytest.raises(ValueError, match="cluster_type must be one of"),
        conn.context(cluster="c"),
    ):
        pass


def test_manifest_not_found_error(
    tmp_path: Path, patch_dependencies: dict[str, Any], monkeypatch: Any
) -> None:
    import geneva.cluster.mgr as cluster_mgr_mod
    import geneva.manifest.mgr as manifest_mgr_mod
    from geneva import connect
    from geneva.cluster import GenevaClusterType

    dummy_cluster = DummyClusterDef(GenevaClusterType.KUBE_RAY)
    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "load", lambda self, name: dummy_cluster
    )
    monkeypatch.setattr(
        manifest_mgr_mod.ManifestConfigManager, "load", lambda self, name: None
    )
    conn = connect(tmp_path)
    with (
        pytest.raises(Exception, match="manifest definition 'm' not found"),
        conn.context(cluster="c", manifest="m"),
    ):
        pass


def test_success_no_manifest(
    tmp_path: Path, patch_dependencies: dict[str, Any], monkeypatch: Any
) -> None:
    import geneva.cluster.mgr as cluster_mgr_mod
    from geneva import connect
    from geneva.cluster import GenevaClusterType
    from geneva.runners.ray.raycluster import ExitMode

    # Prepare dummy cluster and ray cluster
    rc = type("RC", (), {})()
    dummy_cluster = DummyClusterDef(
        GenevaClusterType.KUBE_RAY, use_portforwarding=False
    )
    dummy_cluster.to_ray_cluster = lambda self=dummy_cluster: rc
    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "load", lambda self, name: dummy_cluster
    )
    conn = connect(tmp_path)
    called = patch_dependencies
    with conn.context(cluster="c") as ctx:
        assert ctx == "ctx"
    # Default on_exit should be DELETE
    assert hasattr(rc, "on_exit")
    assert rc.on_exit == ExitMode.DELETE
    # Verify ray_cluster arguments
    assert called == {
        "extra_env": {
            "RAY_BACKEND_LOG_LEVEL": "info",
            "RAY_LOG_TO_DRIVER": "1",
            "RAY_RUNTIME_ENV_LOG_TO_DRIVER_ENABLED": "true",
        },
        "log_to_driver": True,
        "logging_level": 20,
        "manifest": None,
        "ray_cluster": rc,
        "use_portforwarding": False,
    }


def test_success_with_manifest(
    tmp_path: Path, patch_dependencies: dict[str, Any], monkeypatch: Any
) -> None:
    import geneva.cluster.mgr as cluster_mgr_mod
    import geneva.manifest.mgr as manifest_mgr_mod
    from geneva import connect
    from geneva.cluster import GenevaClusterType
    from geneva.runners.ray.raycluster import ExitMode

    rc = type("RC", (), {})()
    dummy_cluster = DummyClusterDef(GenevaClusterType.KUBE_RAY, use_portforwarding=True)
    dummy_cluster.to_ray_cluster = lambda self=dummy_cluster: rc
    dummy_manifest = DummyManifestDef()
    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "load", lambda self, name: dummy_cluster
    )
    monkeypatch.setattr(
        manifest_mgr_mod.ManifestConfigManager,
        "load",
        lambda self, name: dummy_manifest,
    )
    conn = connect(tmp_path)
    called = patch_dependencies
    with conn.context(cluster="c", manifest="m", on_exit=ExitMode.RETAIN) as ctx:
        assert ctx == "ctx"
    assert rc.on_exit == ExitMode.RETAIN
    assert called == {
        "extra_env": {
            "RAY_BACKEND_LOG_LEVEL": "info",
            "RAY_LOG_TO_DRIVER": "1",
            "RAY_RUNTIME_ENV_LOG_TO_DRIVER_ENABLED": "true",
        },
        "log_to_driver": True,
        "logging_level": 20,
        "manifest": dummy_manifest,
        "ray_cluster": rc,
        "use_portforwarding": True,
    }


def test_manifest_image_override(
    tmp_path: Path, patch_dependencies: dict[str, Any], monkeypatch: Any
) -> None:
    import geneva.cluster.mgr as cluster_mgr_mod
    import geneva.manifest.mgr as manifest_mgr_mod
    from geneva import connect

    gc = GenevaClusterBuilder.create("c").build()
    rc = gc.to_ray_cluster()
    manifest = (
        GenevaManifestBuilder.create("m")
        .head_image("head-image")
        .worker_image("worker-image")
        .build()
    )
    monkeypatch.setattr(
        cluster_mgr_mod.ClusterConfigManager, "load", lambda self, name: gc
    )
    monkeypatch.setattr(
        manifest_mgr_mod.ManifestConfigManager,
        "load",
        lambda self, name: manifest,
    )
    conn = connect(tmp_path)
    called = patch_dependencies
    with conn.context(cluster="c", manifest="m") as ctx:
        assert ctx == "ctx"

    # assert the cluster images are set to the manifest images
    rc.head_group.image = "head-image"
    for wg in rc.worker_groups:
        wg.image = "worker-image"
    assert called["ray_cluster"] == rc
